# -*- coding:utf-8 -*- 

"""
Created on 2020/6/15
@author: Meng Lv
@group: Howbuy FDC
@contact: meng.lv@howbuy.com
"""

import pandas as pd
import os
from hbshare.base import vars as ct


def set_token(token):
    df = pd.DataFrame([token], columns=['token'])
    user_home = os.path.expanduser('~')
    fp = os.path.join(user_home, ct.TOKEN_F_P)
    df.to_csv(fp, index=False)
    
    
def get_token():
    user_home = os.path.expanduser('~')
    fp = os.path.join(user_home, ct.TOKEN_F_P)
    if os.path.exists(fp):
        df = pd.read_csv(fp)
        return str(df.iloc[0]['token'])
    else:
        print(ct.TOKEN_ERR_MSG)
        return None

def set_env(env):
    df = pd.DataFrame([env], columns=['env'])
    user_home = os.path.expanduser('~')
    fp = os.path.join(user_home, ct.TOKEN_F_P)
    df.to_csv(fp, index=False)

def get_env():
    user_home = os.path.expanduser('~')
    fp = os.path.join(user_home, ct.ENV_F_P)
    if os.path.exists(fp):
        df = pd.read_csv(fp)
        return str(df.iloc[0]['env'])
    else:
        return None

def is_prod_env():
    return False if ct.ENV_PROD != get_env() else True



#!/usr/bin/python
# coding:utf-8
# 统计基金表现的def

import pandas as pd
import numpy as np
from datetime import datetime
from sqlalchemy import create_engine
import pymysql
from hbshare.quant.CChen.load_data import load_funds_data, load_funds_alpha, load_calendar_extra

pymysql.install_as_MySQLdb()


# 计算每只产品的年化收益，年华波动，夏普，最大回撤等指标
def performance_analysis(
        data_df, risk_free=0.015, start_date=datetime(2019, 1, 1).date(), end_date=datetime.now().date(),
        ret_num_per_year=52, monthly_funds=None, print_info=True
):
    if monthly_funds is None:
        monthly_funds = []
    funds = data_df.columns.tolist()
    date_col = 't_date'
    if 't_date' in funds:
        funds.remove(date_col)
    elif '日期' in funds:
        date_col = '日期'
        funds.remove('日期')
    # data_df['ret'] = data_df['nav'] / data_df['nav'].shift(1)

    result_all = pd.DataFrame()
    dd_all = data_df[[date_col]].copy()
    ret_all = data_df[[date_col]].copy()
    for i in funds:
        single_df = data_df[[date_col, i]]
        # data_before = single_df[single_df[date_col] < start_date][date_col].tolist()
        # if len(data_before) > 0:
        #     start_date_lag1 = data_before[-1]
        # else:
        #     start_date_lag1 = start_date
        single_df = single_df[
            np.array(single_df[date_col] >= start_date) & np.array(single_df[date_col] <= end_date)
            ].reset_index(drop=True)

        single_data = single_df[single_df[i] > 0]
        if len(single_data) > 1:
            if print_info:
                print('\t计算 ' + i + ' 指标')
            start_index = single_data.index[0]
            single_df = single_df.iloc[start_index:].reset_index(drop=True)

            # 剔除最新净值之后的空置（若产品最新净值未按时更新）
            single_df = single_df[single_df[i] > 0].reset_index(drop=True)

            single_df['ret'] = single_df[i] / single_df[i].shift(1) - 1
            single_df['highest'] = single_df[i].cummax()
            single_df['dd'] = (single_df[i] - single_df['highest']) / single_df['highest']
            single_df['ddd'] = single_df['dd'].apply(lambda x: 1 if x < 0 else 0)
            single_df['ddds'] = single_df['ddd'].cumsum()
            single_df['dddg'] = (single_df.index - single_df['ddds']) * single_df['ddd']
            max_dd = min(single_df['dd'])

            dd_days = single_df['dddg'].value_counts().reset_index().sort_values(by='index')
            if len(dd_days) > 1:
                dd_days_hist = max(dd_days['dddg'].tolist()[1:])
                if single_df['ddd'].tolist()[-1] == 0:
                    dd_days_now = 0
                else:
                    dd_days_now = dd_days['dddg'].tolist()[-1]
            else:
                dd_days_hist = np.nan
                dd_days_now = np.nan

            fund_last_days = single_df[date_col][len(single_df) - 1] - single_df[date_col][0]
            cum_return = single_df[i][len(single_df) - 1] / single_df[i][0] - 1
            annualized_return = (single_df[i][len(single_df) - 1] / single_df[i][0]) ** (365 / fund_last_days.days) - 1
            annualized_return_mean = np.mean(single_df['ret'][1:] + 1) ** ret_num_per_year - 1  # 复利年化
            annualized_return_mean_simple = np.mean(single_df['ret'][1:]) * ret_num_per_year
            annualized_vol = np.std(single_df['ret'][1:], ddof=1) * np.sqrt(ret_num_per_year)

            single_df['downside_risk'] = single_df['ret']  #- risk_free
            single_df.loc[
                single_df['downside_risk'] > ((1 + risk_free) ** (1 / ret_num_per_year) - 1),
                'downside_risk'
            ] = 0
            downside_vol = np.std(single_df['downside_risk'][1:], ddof=1) * np.sqrt(ret_num_per_year)
            if annualized_vol == 0:
                sharpe = np.nan
            else:
                sharpe = (annualized_return - risk_free) / annualized_vol
            if downside_vol != 0:
                sortino = (annualized_return - risk_free) / downside_vol
            else:
                sortino = np.nan

            if max_dd < 0:
                calmar = annualized_return / -max_dd
            else:
                calmar = None

            win_rate = sum(single_df['ret'] > 0) / (len(single_df) - 1)

            if len(single_df[single_df['ret'] < 0]['ret']) > 0:
                win_loss = (
                        np.average(single_df[single_df['ret'] > 0]['ret'])
                        / np.average(-single_df[single_df['ret'] < 0]['ret'])
                )
            else:
                win_loss = np.nan

            if i in monthly_funds:
                annualized_return_mean = np.nan
                annualized_vol = np.nan
                sharpe = np.nan
                sortino = np.nan
                win_rate = np.nan
                win_loss = np.nan
        else:
            cum_return = np.nan
            annualized_return = np.nan
            annualized_return_mean = np.nan
            annualized_return_mean_simple = np.nan
            annualized_vol = np.nan
            max_dd = np.nan
            sharpe = np.nan
            sortino = np.nan
            calmar = np.nan
            win_rate = np.nan
            win_loss = np.nan
            dd_days_hist = np.nan
            dd_days_now = np.nan

        result_single = pd.DataFrame(
            {
                start_date.strftime('%Y%m%d') + '以来累计': cum_return,
                start_date.strftime('%Y%m%d') + '以来年化': annualized_return,
                '平均周收益年化': annualized_return_mean,
                # '平均年化收益(单利)': annualized_return_mean_simple,
                '年化波动率': annualized_vol,
                '最大回撤': max_dd,
                '历史最长回撤周': dd_days_hist,
                '已回撤周': dd_days_now,
                'Sharpe': sharpe,
                'Sortino': sortino,
                # '收益峰度': single_df['ret'][1:].kurt(),
                # '收益偏度': single_df['ret'][1:].skew(),
                'Calmar': calmar,
                '投资胜率': win_rate,
                '平均损益比': win_loss

            }, index={i}
        ).T.reset_index()
        if len(result_all) == 0:
            result_all = result_single
        else:
            result_all = result_all.merge(result_single, on='index')

        # if len(dd_all) == 0:
        #     dd_all = single_df[['t_date', 'dd']].rename(columns={'dd': i})
        # else:
        if 'dd' in single_df.columns:
            dd_all = dd_all.merge(single_df[[date_col, 'dd']].rename(columns={'dd': i}), on=date_col, how='left')
        else:
            dd_all[i] = None

        # if len(ret_all) == 0:
        #     ret_all = single_df[['t_date', 'ret']].rename(columns={'ret': i})
        # else:
        if 'ret' in single_df.columns:
            ret_all = ret_all.merge(single_df[[date_col, 'ret']].rename(columns={'ret': i}), on=date_col, how='left')
        else:
            ret_all[i] = None

    return result_all, dd_all, ret_all


def ret(data_df):
    funds = data_df.columns.tolist()
    date_col = 't_date'
    if 't_date' in funds:
        funds.remove(date_col)
    elif '日期' in funds:
        # date_col = '日期'
        funds.remove('日期')
    result_all = data_df[funds].pct_change()
    result_all['t_date'] = data_df['t_date']
    # result_all = data_df['t_date']
    # for i in funds:
    #     single_df = data_df[[date_col, i]].copy()
    #     single_df['ret'] = single_df[i] / single_df[i].shift(1) - 1
    #
    #     result_all = pd.merge(result_all, single_df[['t_date', 'ret']], on='t_date').rename(columns={'ret': i})
    return result_all


# 计算每只产品每月收益
def performance_monthly_ret(data_df):
    # funds = data_df.columns.tolist()
    # date_col = 't_date'
    # if 't_date' in funds:
    #     funds.remove(date_col)
    # elif '日期' in funds:
    #     date_col = '日期'
    #     funds.remove('日期')
    # result_all = pd.DataFrame({'t_date': data_df[date_col][1:]})
    # for i in funds:
    #     ret = data_df[i] / data_df[i].shift(1) - 1
    #     result_all[i] = ret[1:]
    result_all = ret(data_df=data_df)
    return result_all[1:].iloc[::-1]


# 按类型计算产品平均收益率
def performance_ret_per_type(start_date, end_date, keyword='type', freq='w', db_work=''):
    fund_list = pd.read_sql_query(
        'select * from fund_list where class=="cta"', create_engine(db_work)
    )
    type_list = fund_list[keyword].drop_duplicates().tolist()
    type_ret = pd.DataFrame()
    for i in type_list:
        funds = fund_list[fund_list[keyword] == i].reset_index(drop=True)
        funds_data = load_funds_data(
            fund_list=funds,
            first_date=start_date,
            end_date=end_date,
            freq=freq
        )
        funds_ret = performance_monthly_ret(data_df=funds_data)
        funds_ret.to_excel('rwewe.xlsx')
        funds_ret[i] = funds_ret[funds['name'].tolist()].mean(axis=1)
        if len(type_ret) == 0:
            type_ret = funds_ret[['t_date', i]]
        else:
            type_ret = pd.merge(type_ret, funds_ret[['t_date', i]], on='t_date', how='left')

    return type_ret.iloc[::-1]


# 每只产品固定数据
# 收益（本周，近四周，近八周，2020年度，2019年度）
# data_df传入产品净值升序序列
def performance_specific_ret(data_df, form=True, monthly_funds=None, yearly_ret_0=datetime(2017, 12, 29).date()):
    if monthly_funds is None:
        monthly_funds = []
    funds = data_df.columns.tolist()
    date_col = 't_date'
    if 't_date' in funds:
        funds.remove(date_col)
    elif '日期' in funds:
        date_col = '日期'
        funds.remove('日期')

    ret_this_week = []
    ret_last_week = []
    ret_this_4 = []
    ret_this_8 = []
    ret_6_months = []
    ret_2021 = []
    ret_2020 = []  # 年化
    ret_2019 = []  # 年化
    ret_2018 = []  # 年化

    yearly_return = {}

    for i in funds:
        print('\t computing specific index: ' + i)
        single_df = data_df[[date_col, i]].copy()
        # single_df['ret'] = data_df[i] / data_df[i].shift(1)
        if i not in monthly_funds:
            ret_this_week.append(single_df[i][len(single_df) - 1] / single_df[i][len(single_df) - 2] - 1)
            ret_last_week.append(single_df[i][len(single_df) - 2] / single_df[i][len(single_df) - 3] - 1)
            ret_this_4.append(single_df[i][len(single_df) - 1] / single_df[i][len(single_df) - 5] - 1)
            try:
                ret_this_8.append(single_df[i][len(single_df) - 1] / single_df[i][len(single_df) - 9] - 1)
            except KeyError:
                ret_this_8.append(None)

            try:
                ret_6_months.append(single_df[i][len(single_df) - 1] / single_df[i][len(single_df) - 25] - 1)
            except KeyError:
                ret_6_months.append(None)
        else:
            ret_this_week.append(None)
            ret_last_week.append(None)
            ret_this_4.append(None)
            ret_this_8.append(None)
            ret_6_months.append(None)

        single_df['year'] = single_df[date_col].apply(lambda x: x.year)
        range_dates = [max(yearly_ret_0, single_df[date_col][0])]
        year_slice = single_df.drop_duplicates(subset=['year'], keep='last')
        if len(year_slice) > 0:
            range_dates += year_slice[year_slice[date_col] > range_dates[0]][date_col].tolist()

        yearly_return[i] = {}
        for t in range(1, len(range_dates)):
            yearly_data = single_df[
                np.array(single_df[date_col] >= range_dates[t - 1])
                & np.array(single_df[date_col] <= range_dates[t])
            ][i]
            yearly_data = yearly_data[yearly_data > 0].tolist()
            if len(yearly_data) == 0:
                continue
            yearly_return[i][str(range_dates[t].year)] = yearly_data[-1] / yearly_data[0] - 1


        # nav_2018 = single_df[
        #     np.array(single_df[date_col] < datetime(2019, 1, 1).date())
        #     & np.array(single_df[date_col] >= datetime(2017, 12, 29).date())
        # ].reset_index(drop=True)
        # nav_2018 = nav_2018[nav_2018[i] > 0].reset_index(drop=True)
        #
        # if len(nav_2018) > 0:
        #     end_date_2018 = nav_2018[date_col].tolist()[-1]
        # else:
        #     end_date_2018 = datetime(2018, 12, 28).date()
        # nav_2019 = single_df[
        #     np.array(single_df[date_col] < datetime(2020, 1, 1).date())
        #     & np.array(single_df[date_col] >= end_date_2018)
        # ].reset_index(drop=True)
        # nav_2019 = nav_2019[nav_2019[i] > 0].reset_index(drop=True)
        #
        # if len(nav_2019) > 0:
        #     end_date_2019 = nav_2019[date_col].tolist()[-1]
        # else:
        #     end_date_2019 = datetime(2019, 12, 27).date()
        # nav_2020 = single_df[
        #     np.array(single_df[date_col] < datetime(2021, 1, 1).date())
        #     & np.array(single_df[date_col] >= end_date_2019)
        # ].reset_index(drop=True)
        # nav_2020 = nav_2020[nav_2020[i] > 0].reset_index(drop=True)
        #
        # if len(nav_2020) > 0:
        #     end_date_2020 = nav_2020[date_col].tolist()[-1]
        # else:
        #     end_date_2020 = datetime(2020, 12, 31).date()
        # nav_2021 = single_df[
        #     np.array(single_df[date_col] >= end_date_2020)
        # ].reset_index(drop=True)
        # nav_2021 = nav_2021[nav_2021[i] > 0].reset_index(drop=True)
        #
        # if len(nav_2018) > 1:
        #     ret_2018.append(nav_2018[i].tolist()[-1] / nav_2018[nav_2018[i] > 0][i].tolist()[0] - 1)
        # else:
        #     ret_2018.append(None)
        #
        # if len(nav_2019) > 1:
        #     ret_2019.append(nav_2019[i].tolist()[-1] / nav_2019[nav_2019[i] > 0][i].tolist()[0] - 1)
        # else:
        #     ret_2019.append(None)
        #
        # if len(nav_2020) > 1:
        #     ret_2020.append(nav_2020[i].tolist()[-1] / nav_2020[nav_2020[i] > 0][i].tolist()[0] - 1)
        # else:
        #     ret_2020.append(None)
        #
        # if len(nav_2021) > 1:
        #     ret_2021.append(nav_2021[i].tolist()[-1] / nav_2021[nav_2021[i] > 0][i].tolist()[0] - 1)
        # else:
        #     ret_2021.append(None)

        # try:
        #     ret_2021.append(
        #         nav_2021[i].tolist()[-1] / nav_2021[nav_2021[i] > 0][i].tolist()[0] - 1
        #     )
        # except:
        #     ret_2021.append(None)
        #
        # try:
        #     fund_2020_days = nav_2020[date_col][len(nav_2020) - 1] - nav_2020[date_col][0]
        #     ret_2020.append(
        #         (nav_2020[i].tolist()[-1] / nav_2020[nav_2020[i] > 0][i].tolist()[0]) ** (365 / fund_2020_days.days) - 1
        #     )
        # except IndexError:
        #     ret_2020.append(None)
        #
        # try:
        #     fund_2019_days = nav_2019[date_col][len(nav_2019) - 1] - nav_2019[date_col][0]
        #     if fund_2019_days.days < 90:
        #         ret_2019.append(None)
        #     else:
        #         ret_2019.append(
        #             (
        #                     nav_2019[i].tolist()[-1] / nav_2019[nav_2019[i] > 0][i].tolist()[0]
        #             ) ** (365 / fund_2019_days.days) - 1
        #         )
        # except:
        #     ret_2019.append(None)
        #
        # if len(nav_2018) > 1:
        #     fund_2018_days = nav_2018[date_col][len(nav_2018) - 1] - nav_2018[date_col][0]
        #     # 运行超过半年才计算年化
        #     if fund_2018_days.days < 90:
        #         ret_2018.append(None)
        #     else:
        #         ret_2018.append(
        #             (
        #                     nav_2018[i].tolist()[-1] / nav_2018[nav_2018[i] > 0][i].tolist()[0]
        #             ) ** (365 / fund_2018_days.days) - 1
        #         )
        # else:
        #     ret_2018.append(None)

    if form:
        result_df = pd.DataFrame(
            {
                '本周': ret_this_week,
                '上周': ret_last_week,
                '近四周': ret_this_4,
                '近八周': ret_this_8,
                '近六月': ret_6_months,
                # '2021': ret_2021,
                # '2020': ret_2020,
                # '2019': ret_2019,
                # '2018': ret_2018
            }, index=funds
        ).T.reset_index()

        return result_df.append(pd.DataFrame(yearly_return).reset_index().sort_values('index', ascending=False))

    else:
        result_df = pd.DataFrame(
            {
                '本周收益': ret_this_week,
                '上周收益': ret_last_week,
                '近四周收益': ret_this_4
            }, index=funds
        ).sort_values(by='本周收益', ascending=True)

        return result_df.T.reset_index()


# 统计汇总
def update_performance(start_d, end_d, funds, db_path, cal_db_path, month_release, data_type=''):
    fund_list = pd.read_sql_query(
        'select * from fund_list where `name` in ' + str(tuple(funds)), create_engine(db_path)
    )

    nav_origin = None

    if data_type.lower() != 'alpha':
        funds_data = load_funds_data(
            fund_list=fund_list,
            first_date=datetime(2000, 1, 1).date(),
            end_date=end_d,
            db_path=db_path,
            cal_db_path=cal_db_path,
            print_info=True
        )
        funds_data_monthly = load_funds_data(
            fund_list=fund_list,
            first_date=start_d,
            end_date=end_d,
            freq='month',
            db_path=db_path,
            cal_db_path=cal_db_path,
            print_info=True
        )

    else:
        funds_data = load_funds_alpha(
            fund_list=fund_list,
            first_date=datetime(2000, 1, 1).date(),
            end_date=end_d,
            db_path=db_path,
            cal_db_path=cal_db_path,
            print_info=True
        )
        nav_origin = funds_data['nav']
        funds_data = funds_data['eav']

        cal_daily = load_calendar_extra(freq='', end_date=end_d, db_path=cal_db_path)
        cal_daily = pd.merge(cal_daily, funds_data, on='t_date', how='left')
        cal_daily = cal_daily.fillna(method='ffill')

        funds_data_monthly = pd.merge(
            load_calendar_extra(freq='month', end_date=end_d, db_path=cal_db_path),
            cal_daily,
            on='t_date', how='left'
        )
        funds_data_monthly = funds_data_monthly[funds_data_monthly['t_date'] >= start_d]

    result_monthly_ret = ret(funds_data_monthly)[1:].iloc[::-1]
    result_key, result_dd, result_ret = performance_analysis(
        funds_data[funds_data['t_date'] >= start_d].reset_index(drop=True), monthly_funds=month_release
    )

    result_key2, _, _ = performance_analysis(
        funds_data[funds_data['t_date'] >= start_d].reset_index(drop=True), monthly_funds=month_release,
        start_date=funds_data['t_date'].tolist()[-52]
    )

    result_specific = performance_specific_ret(
        funds_data[funds_data['t_date'] >= start_d].reset_index(drop=True), monthly_funds=month_release
    )

    fund_name_list = []
    fund_date_list = []
    fund_start_date = []
    for n in fund_list['name']:
        fund_name_list.append(n)
        try:
            fund_date_list.append(funds_data[funds_data[n] > 0]['t_date'].tolist()[-1].strftime('%Y-%m-%d'))
        except IndexError:
            fund_date_list.append(None)

        try:
            fund_start_date.append(funds_data[funds_data[n] > 0]['t_date'].tolist()[0].strftime('%Y/%m/%d'))
        except IndexError:
            fund_start_date.append(None)

    result_date = pd.DataFrame(
        {
            '起始日期': fund_start_date,
            '最新日期': fund_date_list
        }, index=fund_name_list
    ).T.reset_index()

    results_all = result_date.append(result_specific).append(result_key2).append(result_key)

    nav_all = funds_data

    dd_all = result_dd

    weekly_ret_all = result_ret

    monthly_ret_all = result_monthly_ret

    nav_all_origin = nav_origin

    nav_2019 = nav_all[np.array(nav_all['t_date'] >= datetime(2018, 12, 28).date())].copy()
    nav_2020 = nav_all[np.array(nav_all['t_date'] >= datetime(2019, 12, 27).date())].copy()

    funds = nav_all.columns.tolist()
    funds.remove('t_date')

    for i in funds:
        try:
            nav_2019.loc[:, i] = nav_2019[i] / nav_2019[nav_2019[i] > 0][i].tolist()[0]
        except IndexError:
            nav_2019.loc[:, i] = None

        try:
            nav_2020.loc[:, i] = nav_2020[i] / nav_2020[nav_2020[i] > 0][i].tolist()[0]
        except IndexError:
            nav_2020.loc[:, i] = None

    result = {
        'main': results_all.reset_index(drop=True),
        'nav2020': nav_2020.sort_values(by='t_date', ascending=False).reset_index(drop=True),
        'nav2019': nav_2019.sort_values(by='t_date', ascending=False).reset_index(drop=True),
        'nav': nav_all.sort_values(by='t_date', ascending=False).reset_index(drop=True),
        'dd': dd_all.sort_values(by='t_date', ascending=False).reset_index(drop=True),
        'weekly_ret': weekly_ret_all.sort_values(by='t_date', ascending=False).reset_index(drop=True),
        'monthly_ret': monthly_ret_all
    }
    if data_type == 'alpha':
        result['nav_o'] = nav_all_origin.sort_values(by='t_date', ascending=False).reset_index(drop=True)
    return result




"""
线上和本地的风险模型结果校验程序
"""
import pandas as pd
import hbshare as hbs
import matplotlib.pyplot as plt
import numpy as np
import os

path = r'D:\kevin\risk_model_jy\RiskModel\data'
style_name_list = [
    "size", "beta", "momentum", "earnyield", "resvol", "growth", "btop", "leverage", "liquidity", "sizenl"]


def style_factor_check(date, factor_name):
    sql_script = "SELECT * FROM st_ashare.r_st_barra_style_factor where TRADE_DATE = '{}'".format(date)
    res = hbs.db_data_query('alluser', sql_script, page_size=5000)
    style_factor_ol = pd.DataFrame(res['data'])[['ticker'] + style_name_list].set_index('ticker')

    data_path = os.path.join(path, r'zzqz_sw\style_factor')
    style_factor_local = pd.read_csv(
        os.path.join(data_path, '{0}.csv'.format(date)), dtype={"ticker": str}).set_index('ticker')

    factor_compare = pd.merge(style_factor_ol[factor_name].to_frame(name=factor_name + 'ol'),
                              style_factor_local[factor_name].to_frame(name=factor_name + 'local'),
                              left_index=True, right_index=True)
    print(factor_compare)


def factor_return_check(date):
    sql_script = "SELECT * FROM st_ashare.r_st_barra_factor_return where TRADE_DATE = '{}'".format(date)
    res = hbs.db_data_query('alluser', sql_script, page_size=5000)
    factor_return_ol = pd.DataFrame(res['data']).set_index('factor_name')['factor_ret']

    data_path = os.path.join(path, r'zzqz_sw\factor_return')
    factor_return_local = pd.read_csv(
        os.path.join(data_path, '{0}.csv'.format(date)), dtype={"ticker": str}).set_index('factor_name')['factor_ret']

    return_compare = pd.merge(factor_return_ol.to_frame('f_ret_ol'), factor_return_local.to_frame('f_ret_local'),
                              left_index=True, right_index=True)
    print(return_compare)


def specific_return_check(date):
    sql_script = "SELECT * FROM st_ashare.r_st_barra_specific_return where TRADE_DATE = '{}'".format(date)
    res = hbs.db_data_query('alluser', sql_script, page_size=5000)
    s_ret_ol = pd.DataFrame(res['data'])
    s_ret_ol['ticker'] = s_ret_ol['ticker'].apply(lambda x: x.zfill(6))
    s_ret_ol = s_ret_ol.set_index('ticker')['s_ret']

    data_path = os.path.join(path, r'zzqz_sw\specific_return')
    s_ret_local = pd.read_csv(
        os.path.join(data_path, '{0}.csv'.format(date)), dtype={"ticker": str}).set_index('ticker')['s_ret']

    s_ret_compare = pd.merge(s_ret_ol.to_frame('s_ret_ol'), s_ret_local.to_frame('s_ret_local'),
                             left_index=True, right_index=True)
    print(s_ret_compare)


def factor_cov_check(date):
    data_path = os.path.join(path, r'zzqz_sw\factor_return')
    factor_return_local = pd.read_csv(
        os.path.join(data_path, '{0}.csv'.format(date)), dtype={"ticker": str}).set_index('factor_name')['factor_ret']
    name_list = [x.lower() for x in factor_return_local.index.tolist()]

    sql_script = "SELECT * FROM st_ashare.r_st_barra_factor_cov where TRADE_DATE = '{}'".format(date)
    res = hbs.db_data_query('alluser', sql_script, page_size=5000)
    factor_cov_ol = pd.DataFrame(res['data']).set_index('factor_name').reindex(factor_return_local.index)[name_list]

    data_path = os.path.join(path, r'zzqz_sw\factor_cov\short')
    factor_cov_local = pd.read_csv(os.path.join(data_path, '{0}.csv'.format(date))).set_index('factor_name')
    factor_cov_local = factor_cov_local.reindex(factor_return_local.index)[factor_return_local.index]

    dm, um = np.linalg.eig(factor_cov_ol)
    dm1, um1 = np.linalg.eig(factor_cov_local)
    dm_df = pd.DataFrame({"dm_ol": dm, "dm_local": dm1})
    dm_df.plot.bar()
    plt.show()


def srisk_check(date):
    sql_script = "SELECT * FROM st_ashare.r_st_barra_s_risk where TRADE_DATE = '{}'".format(date)
    res = hbs.db_data_query('alluser', sql_script, page_size=5000)
    srisk_ol = pd.DataFrame(res['data']).set_index('ticker')['s_ret']

    data_path = os.path.join(path, r'zzqz_sw\srisk\short')
    srisk_local = pd.read_csv(
        os.path.join(data_path, '{0}.csv'.format(date)), dtype={"ticker": str}).set_index('ticker')

    srisk_compare = pd.merge(srisk_ol.to_frame('srisk_ol'), srisk_local, left_index=True, right_index=True)
    print(srisk_compare)


if __name__ == '__main__':
    trade_date = '20201201'
    # style_factor_check(trade_date, 'growth')
    factor_return_check(trade_date)
    # specific_return_check(trade_date)
    # factor_cov_check(trade_date)
    # srisk_check(trade_date)


"""
主观股多的估值表分析
"""
import os
import datetime
import pandas as pd
import numpy as np
import hbshare as hbs
from hbshare.quant.Kevin.asset_allocation.macro_index.util import create_table, delete_duplicate_records, WriteToDB
from sqlalchemy import create_engine
from hbshare.quant.Kevin.rm_associated.config import engine_params
from hbshare.fe.common.util.config import industry_name, industry_cluster_dict
import plotly
import plotly.graph_objs as go
import pyecharts.options as opts
from pyecharts.charts import Bar, Timeline
import plotly.figure_factory as ff
from hbshare.db.simu import valuation
from hbshare.fe.XZ import  db_engine

# from plotly.offline import plot as plot_ly

# plotly.offline.init_notebook_mode(connected=True)

hbdb=db_engine.HBDB()

def plot_render(plot_dic, width=1200, height=600, **kwargs):
    kwargs['output_type'] = 'div'
    plot_str = plotly.offline.plot(plot_dic, **kwargs)
    print('%%angular <div style="height: %ipx; width: %spx"> %s </div>' % (height, width, plot_str))

# def plot_render(plot_dic, width=1200, height=600, **kwargs):
#     data=plot_dic['data']
#     layout=plot_dic['layout']
#     fig = go.Figure(data=data, layout=layout)
#     fig.show()

class HoldingExtractor:
    def __init__(self, table_name,fund_name,is_increment=1,data_path=None,fund_code=None,date_list=None):
        self.data_path = data_path
        self.table_name = table_name
        self.fund_name = fund_name
        self.is_increment = is_increment
        self.fund_code=fund_code
        self.date_list=date_list
        self.total_asset_dmlist=['资产类合计:','资产合计']
        self.net_asset_dmlist=['基金资产净值:','资产净值','资产资产净值:']
        self.debt_dmlist=['负债类合计:']
        self.ashare_dmlist=['11020101','1102.01.01.','1101010101']
        self.szshare_dmlist=['11023101','1102.33.01.','1101013101']
        self.cyb_dmlistt=['11024101','1101014101']
        self.kcb_dmlist=['1102C101','1101C101']
        self.jjtz_dmlist = ['110104', '1105']
        if(data_path is not None):
            self._load_portfolio_weight()

    @staticmethod
    def _shift_date(date):
        trade_dt = datetime.datetime.strptime(date, '%Y%m%d')
        pre_date = (trade_dt - datetime.timedelta(days=30)).strftime('%Y%m%d')

        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            pre_date, date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)

        trading_day_list = df[df['isOpen'] == 1]['calendarDate'].tolist()

        return trading_day_list[-1]

    def valuation_data2DB(self,jjdm,date_list,fund_name):
        portfolio_weight_list = []
        for date in date_list:
            data=valuation.get_prod_valuation_by_jjdm_gzrq(jjdm,date)
            if(data.empty):
                print("data for {0} missed,continue...".format(date))
                continue
            date = self._shift_date(str(date))
            columns_map=dict(zip(['kmdm','sz','kmmc'],['科目代码','市值','科目名称']))

            data.rename(columns=columns_map,inplace=True)

            for dm in self.total_asset_dmlist:
                if(len(data[data['科目代码'] == dm])>0):
                    total_asset = float(data[data['科目代码'] ==dm]['市值'].values[0])
                    break

            net_asset=0
            for dm in self.net_asset_dmlist:
                if(len(data[data['科目代码'] == dm])>0):
                    net_asset = float(data[data['科目代码'] == dm]['市值'].values[0])
                    break

            if(net_asset==0):
                for dm in self.debt_dmlist:
                    if (len(data[data['科目代码'] == dm]) > 0):
                        debt = float(data[data['科目代码'] == dm]['市值'].values[0])
                        net_asset=total_asset-debt
                        break

            leverage = total_asset / net_asset
            # A股
            sh=pd.DataFrame()
            for dm in self.ashare_dmlist:
                if(len(data[data['科目代码'].str.startswith(dm)])>0):
                    sh = data[data['科目代码'].str.startswith(dm)]
                    break
            # sh = data[data['科目代码'].str.startswith('11020101')]
            sz=pd.DataFrame()
            for dm in self.szshare_dmlist:
                if(len(data[data['科目代码'].str.startswith(dm)])>0):
                    sz = data[data['科目代码'].str.startswith(dm)]
                    break
            cyb = pd.DataFrame()
            for dm in self.cyb_dmlistt:
                if (len(data[data['科目代码'].str.startswith(dm)]) > 0):
                    cyb = data[data['科目代码'].str.startswith(dm)]
                    break
            kcb = pd.DataFrame()
            for dm in self.kcb_dmlist:
                if (len(data[data['科目代码'].str.startswith(dm)]) > 0):
                    kcb = data[data['科目代码'].str.startswith(dm)]
                    break

            if(len(sh)==0 and len(sz)==0 and len(cyb)==0 and len(kcb)==0):
                print('No stock exist for {0} at {1}..'.format(fund_name,date))
                continue

            equity_a = pd.concat([sh, sz, cyb, kcb], axis=0)

            equity_a['len'] = equity_a['科目代码'].apply(lambda x: len(x))
            equity_a = equity_a[equity_a['len'] > 10]

            if("." not in equity_a['科目代码'].values[0]):
                equity_a['ticker'] = equity_a['科目代码'].apply(lambda x: x[-6:])
            else:
                equity_a['ticker'] = equity_a['科目代码'].apply(lambda x: x[-8:-2])

            equity_a['weight'] = equity_a['市值'].astype(float) / net_asset
            equity_a = equity_a.rename(columns={"科目名称": "sec_name"})[['ticker', 'sec_name', 'weight']]
            # 港股
            hk1 = data[data['科目代码'].str.startswith('11028101')]
            hk2 = data[(data['科目代码'].str.startswith('11028201')) | (data['科目代码'].str.startswith('11028301'))]
            equity_hk = pd.concat([hk1, hk2], axis=0).fillna('')
            equity_hk['len'] = equity_hk['科目代码'].apply(lambda x: len(x))
            equity_hk = equity_hk[equity_hk['len'] > 8]
            equity_hk['ticker'] = equity_hk['科目代码'].apply(lambda x: x[-6:])
            equity_hk['科目名称'] = equity_hk['科目名称'].apply(lambda x: x.strip())  # add
            equity_hk['weight'] = equity_hk['市值'].astype(float) / net_asset
            equity_hk = equity_hk.rename(columns={"科目名称": "sec_name"})[['ticker', 'sec_name', 'weight']]
            equity_hk = equity_hk.groupby(['sec_name', 'ticker'])['weight'].sum().reset_index()
            # 债券
            tmp = data[data['科目代码'] == '1103']
            if tmp.empty:
                bond_df = pd.DataFrame()
            else:
                bond_ratio = tmp['市值'].astype(float).values[0] / net_asset
                bond_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
                bond_df.loc[0] = ['b00001', '债券投资', bond_ratio]
            # 基金
            for dm in self.jjtz_dmlist:
                if (len(data[data['科目代码']==dm]) > 0):
                    tmp = data[data['科目代码'] == dm]
                    break
            if tmp.empty:
                fund_df = pd.DataFrame()
            else:
                fund_ratio = tmp['市值'].astype(float).values[0] / net_asset
                fund_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
                fund_df.loc[0] = ['f00001', '基金投资', fund_ratio]

            df = pd.concat([equity_a, equity_hk, bond_df, fund_df], axis=0)
            # 其他类
            cash_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
            cash_df.loc[0] = ['c00001', '现金类投资', leverage - df['weight'].sum()]
            df = pd.concat([df, cash_df], axis=0)
            df['trade_date'] = date
            if(len(df[df['ticker'].str[0].isin(['0', '3', '6'])])>0):
                portfolio_weight_list.append(df)

        portfolio_weight_df = pd.concat(portfolio_weight_list)
        portfolio_weight_df['fund_name'] = fund_name

        return  portfolio_weight_df

    def _load_portfolio_weight(self):
        filenames = os.listdir(self.data_path)
        filenames = [x for x in filenames if x.split('.')[-1] in ['xls', 'xlsx']]

        portfolio_weight_list = []
        for file_name in filenames:
            if self.fund_name == '亘曦2号':
                date = file_name.split('.')[0].split('_')[-1][:-3].replace('-', '')
            elif self.fund_name in ['富乐一号', '仁布积极进取1号']:
                date = file_name.split('.')[0].split('_')[-2]
            else:
                date = file_name.split('.')[0].split('_')[-1]
            # shift_date
            date = self._shift_date(date)
            data = pd.read_excel(
                os.path.join(self.data_path, file_name), sheet_name=0, header=3).dropna(subset=['科目代码'])
            net_asset = data[data['科目代码'] == '基金资产净值:']['市值'].values[0]
            total_asset = data[data['科目代码'] == '资产类合计:']['市值'].values[0]
            leverage = total_asset / net_asset
            # A股
            sh = data[data['科目代码'].str.startswith('11020101')]
            sz = data[data['科目代码'].str.startswith('11023101')]
            cyb = data[data['科目代码'].str.startswith('11024101')]
            kcb = data[data['科目代码'].str.startswith('1102C101')]
            equity_a = pd.concat([sh, sz, cyb, kcb], axis=0)
            # sh1 = data[data['科目代码'].str.startswith('11021101')]
            # sz1 = data[data['科目代码'].str.startswith('11021201')]
            # cyb1 = data[data['科目代码'].str.startswith('11021501')]
            # kcb1 = data[data['科目代码'].str.startswith('1102D201')]
            # xsg = data[data['科目代码'].str.startswith('11028401')]
            # equity_cr = pd.concat([sh1, sz1, cyb1, kcb1, xsg], axis=0)
            # equity_a = pd.concat([equity_a, equity_cr], axis=0)
            equity_a['len'] = equity_a['科目代码'].apply(lambda x: len(x))
            equity_a = equity_a[equity_a['len'] > 8]
            equity_a['ticker'] = equity_a['科目代码'].apply(lambda x: x[-6:])
            equity_a['weight'] = equity_a['市值'] / net_asset
            equity_a = equity_a.rename(columns={"科目名称": "sec_name"})[['ticker', 'sec_name', 'weight']]
            # 港股
            hk1 = data[data['科目代码'].str.startswith('11028101')]
            hk2 = data[(data['科目代码'].str.startswith('11028201')) | (data['科目代码'].str.startswith('11028301'))]
            equity_hk = pd.concat([hk1, hk2], axis=0)
            equity_hk['len'] = equity_hk['科目代码'].apply(lambda x: len(x))
            equity_hk = equity_hk[equity_hk['len'] > 8]
            equity_hk['ticker'] = equity_hk['科目代码'].apply(lambda x: x[-6:])
            equity_hk['科目名称'] = equity_hk['科目名称'].apply(lambda x: x.strip())  # add
            equity_hk['weight'] = equity_hk['市值'] / net_asset
            equity_hk = equity_hk.rename(columns={"科目名称": "sec_name"})[['ticker', 'sec_name', 'weight']]
            equity_hk = equity_hk.groupby(['sec_name', 'ticker'])['weight'].sum().reset_index()
            # 债券
            tmp = data[data['科目代码'] == '1103']
            if tmp.empty:
                bond_df = pd.DataFrame()
            else:
                bond_ratio = tmp['市值'].values[0] / net_asset
                bond_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
                bond_df.loc[0] = ['b00001', '债券投资', bond_ratio]
            # 基金
            tmp = data[data['科目代码'] == '1105']
            if tmp.empty:
                fund_df = pd.DataFrame()
            else:
                fund_ratio = tmp['市值'].values[0] / net_asset
                fund_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
                fund_df.loc[0] = ['f00001', '基金投资', fund_ratio]

            df = pd.concat([equity_a, equity_hk, bond_df, fund_df], axis=0)
            # 其他类
            cash_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
            cash_df.loc[0] = ['c00001', '现金类投资', leverage - df['weight'].sum()]
            df = pd.concat([df, cash_df], axis=0)
            df['trade_date'] = date
            portfolio_weight_list.append(df)

        portfolio_weight_df = pd.concat(portfolio_weight_list)
        portfolio_weight_df['fund_name'] = self.fund_name

        return portfolio_weight_df

    def Pre_test(self,fromdb=False):

        if (fromdb):
            data = self.valuation_data2DB(self.fund_code, self.date_list, self.fund_name)
        else:
            data = self._load_portfolio_weight()

        start_date=data['trade_date'].unique().tolist()[0]
        end_date=data['trade_date'].unique().tolist()[-1]

        HoldingAnalysor(self.fund_name, start_date=start_date, end_date=end_date,inputdf=data).get_construct_result()

    def writeToDB(self,fromdb=False):
        if self.is_increment == 1:
            if(fromdb):
                data=self.valuation_data2DB(self.fund_code,self.date_list,self.fund_name)
            else:
                data = self._load_portfolio_weight()
            trading_day_list = data['trade_date'].unique().tolist()
            sql_script = "delete from {} where trade_date in ({}) and fund_name = '{}'".format(
                self.table_name, ','.join(trading_day_list), self.fund_name)
            # delete first
            delete_duplicate_records(sql_script)
            # add new records

            WriteToDB().write_to_db(data, self.table_name)
        else:
            sql_script = """
                create table {}(
                id int auto_increment primary key,
                trade_date date not null,
                ticker varchar(10),
                sec_name varchar(20),
                weight decimal(5, 4),
                fund_name varchar(40))
            """.format(self.table_name)
            create_table(self.table_name, sql_script)
            data = self._load_portfolio_weight()
            WriteToDB().write_to_db(data, self.table_name)

class HoldingAnalysor:
    def __init__(self, fund_name, start_date, end_date, threshold=1,inputdf=None):
        self.fund_name = fund_name
        self.start_date = start_date
        self.end_date = end_date
        self.threshold = threshold
        if(inputdf is None):
            self._load_data()
        else:self._load_data(inputdf)

    def _load_portfolio_weight(self):
        # sql_script = "SELECT * FROM subjective_fund_holding where fund_name = '{}' and " \
        #              "trade_date >= {} and trade_date <= {}".format(self.fund_name, self.start_date, self.end_date)
        # engine = create_engine(engine_params)
        # holding_df = pd.read_sql(sql_script, engine)
        # holding_df['trade_date'] = holding_df['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))


        sql="select * from st_hedge.r_st_sm_subjective_fund_holding where jjdm='{0}' and trade_date >= '{1}' and trade_date <= '{2}' "\
            .format(self.fund_name,self.start_date,self.end_date)
        holding_df=hbdb.db2df(sql,db='highuser')
        holding_df['trade_date']=holding_df['trade_date'].astype(str)

        return holding_df[['trade_date', 'ticker', 'sec_name', 'weight']]

    @staticmethod
    def _load_shift_date(date):
        trade_dt = datetime.datetime.strptime(date, '%Y%m%d')
        pre_date = (trade_dt - datetime.timedelta(days=100)).strftime('%Y%m%d')

        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            pre_date, date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)

        trading_day_list = df[df['isMonthEnd'] == 1]['calendarDate'].tolist()

        return trading_day_list[-1]

    @staticmethod
    def _load_benchmark_weight(benchmark_id, shift_date, date):
        sql_script = "SELECT * FROM hsjy_gg.SecuMain where SecuCategory = 4 and SecuCode = '{}'".format(benchmark_id)
        res = hbs.db_data_query('readonly', sql_script)
        index_info = pd.DataFrame(res['data'])
        inner_code = index_info.set_index('SECUCODE').loc[benchmark_id, 'INNERCODE']

        sql_script = "SELECT (select a.SecuCode from hsjy_gg.SecuMain a where a.InnerCode = b.InnerCode and " \
                     "rownum = 1) SecuCode, b.EndDate, b.Weight FROM hsjy_gg.LC_IndexComponentsWeight b WHERE " \
                     "b.IndexCode = '{}' and b.EndDate = to_date('{}', 'yyyymmdd')".format(inner_code, shift_date)
        data = pd.DataFrame(hbs.db_data_query('readonly', sql_script)['data'])
        weight_df = data.rename(
            columns={"SECUCODE": "ticker", "ENDDATE": "effDate", "WEIGHT": "weight"})
        weight_df['benchmark_id'] = benchmark_id
        weight_df['trade_date'] = date

        return weight_df[['trade_date', 'ticker', 'benchmark_id']]

    @staticmethod
    def _load_security_sector(portfolio_weight):
        equity_portfolio_weight = portfolio_weight[portfolio_weight['ticker'].str[0].isin(['0', '3', '6'])]
        trading_day_list = sorted(portfolio_weight['trade_date'].unique())
        cols_list = ['ticker'] + [x.lower() for x in industry_name['sw'].values()]
        security_sector_list = []
        for date in trading_day_list:
            ticker_list = equity_portfolio_weight[equity_portfolio_weight['trade_date'] == date]['ticker'].tolist()
            sql_script = "SELECT {} FROM st_ashare.r_st_barra_style_factor where TRADE_DATE = '{}' and " \
                         "ticker in ({})".format(','.join(cols_list), date,
                                                 ','.join("'{0}'".format(x) for x in ticker_list))
            res = hbs.db_data_query('alluser', sql_script, page_size=5000)
            data = pd.DataFrame(res['data']).set_index('ticker')
            reverse_ind = dict([(value.lower(), key) for (key, value) in industry_name['sw'].items()])
            ind_exposure = data[reverse_ind.keys()].rename(columns=reverse_ind)
            ind_exposure = ind_exposure.reset_index().melt(
                id_vars=['ticker'], value_vars=list(reverse_ind.values()), var_name='industryName1', value_name='sign')
            ind_exposure = ind_exposure[ind_exposure['sign'] == '1']
            ind_exposure['trade_date'] = date
            ind_exposure = ind_exposure.set_index('ticker').reindex(ticker_list).dropna().reset_index()
            security_sector_list.append(ind_exposure)

        security_sector_df = pd.concat(security_sector_list)[['trade_date', 'ticker', 'industryName1']]

        return security_sector_df

    @staticmethod
    def _load_security_value(portfolio_weight):
        equity_portfolio_weight = portfolio_weight[portfolio_weight['ticker'].str[0].isin(['0', '3', '6'])]
        trading_day_list = sorted(portfolio_weight['trade_date'].unique())
        value_data_list = []
        for date in trading_day_list:
            ticker_list = equity_portfolio_weight[equity_portfolio_weight['trade_date'] == date]['ticker'].tolist()
            # 主板
            sql_script = "SELECT PE, PB, DividendRatio, TotalMV, SecuCode FROM " \
                         "(SELECT b.PE, b.PB, b.DividendRatio, b.TotalMV, a.SecuCode, " \
                         "row_number() over(partition by a.InnerCode order by a.InnerCode) rn FROM " \
                         "hsjy_gg.LC_DIndicesForValuation b join hsjy_gg.SecuMain a on a.InnerCode = b.InnerCode and " \
                         "a.SecuMarket in (83, 90) and a.SecuCategory = 1 WHERE " \
                         "b.TradingDay = to_date('{}', 'yyyymmdd') and a.SecuCode in ({})) " \
                         "WHERE rn = 1".format(date, ','.join("'{0}'".format(x) for x in ticker_list))
            res = hbs.db_data_query('readonly', sql_script, page_size=5000)
            data_main = pd.DataFrame(res['data']).rename(columns={"PE": "PETTM"})
            # 科创板
            sql_script = "SELECT PETTM, PB, DividendRatio, TotalMV, SecuCode FROM " \
                         "(SELECT b.PETTM, b.PB, b.DividendRatio, b.TotalMV, a.SecuCode, " \
                         "row_number() over(partition by a.InnerCode order by a.InnerCode) rn FROM " \
                         "hsjy_gg.LC_STIBDIndiForValue b join hsjy_gg.SecuMain a on a.InnerCode = b.InnerCode and " \
                         "a.SecuMarket in (83, 90) and a.SecuCategory = 1 WHERE " \
                         "b.TradingDay = to_date('{}', 'yyyymmdd') and a.SecuCode in ({})) " \
                         "WHERE rn = 1".format(date, ','.join("'{0}'".format(x) for x in ticker_list))
            res = hbs.db_data_query('readonly', sql_script, page_size=5000)
            data_stib = pd.DataFrame(res['data'])

            data = pd.concat([data_main, data_stib]).rename(columns={"SECUCODE": "ticker", "TOTALMV": "MarketValue"})
            data['MarketValue'] /= 1e+8
            del data['ROW_ID']
            data = data.dropna(subset=['ticker'])
            data['trade_date'] = date
            value_data_list.append(data)

        security_value_df = pd.concat(value_data_list)

        return security_value_df

    def _load_data(self,inputdf=None):
        if(inputdf is None):
            portfolio_weight_df = self._load_portfolio_weight()
        else:
            portfolio_weight_df=inputdf
        date_list = sorted(portfolio_weight_df['trade_date'].unique())
        benchmark_weight = []
        for date in date_list:
            shift_date = self._load_shift_date(date)
            weight_300 = self._load_benchmark_weight('000300', shift_date, date)
            weight_500 = self._load_benchmark_weight('000905', shift_date, date)
            weight_1000 = self._load_benchmark_weight('000852', shift_date, date)
            benchmark_weight.append(pd.concat([weight_300, weight_500, weight_1000]))

        benchmark_weight = pd.concat(benchmark_weight)

        security_sector_df = self._load_security_sector(portfolio_weight_df)

        security_value_df = self._load_security_value(portfolio_weight_df)

        self.data_param = {"portfolio_weight": portfolio_weight_df,
                           "benchmark_weight": benchmark_weight,
                           "security_sector_df": security_sector_df,
                           "security_value_df": security_value_df}

    @staticmethod
    def _calculate_asset_allo_series(portfolio_weight):
        date_list = sorted(portfolio_weight['trade_date'].unique())
        equity_a_series = portfolio_weight[portfolio_weight['ticker'].str[0].isin(['0', '3', '6'])].groupby(
            'trade_date')['weight'].sum().reindex(date_list).fillna(0.).to_frame('A股')
        equity_hk_series = portfolio_weight[portfolio_weight['ticker'].str.startswith('H')].groupby(
            'trade_date')['weight'].sum().reindex(date_list).fillna(0.).to_frame('港股')
        bond_series = portfolio_weight[portfolio_weight['ticker'].str.startswith('b')].groupby(
            'trade_date')['weight'].sum().reindex(date_list).fillna(0.).to_frame('债券')
        fund_series = portfolio_weight[portfolio_weight['ticker'].str.startswith('f')].groupby(
            'trade_date')['weight'].sum().reindex(date_list).fillna(0.).to_frame('基金')
        cash_series = portfolio_weight[portfolio_weight['ticker'].str.startswith('c')].groupby(
            'trade_date')['weight'].sum().reindex(date_list).fillna(0.).to_frame('现金类')
        asset_allo_series = pd.concat(
            [equity_a_series, equity_hk_series, bond_series, fund_series, cash_series], axis=1)
        # asset_allo_series['现金类'] = 1 - asset_allo_series.sum(axis=1)

        return asset_allo_series

    @staticmethod
    def _calculate_ind_allo_series(portfolio_weight, security_sector_df):
        weight_df = pd.merge(portfolio_weight, security_sector_df, on=['trade_date', 'ticker'])
        grouped_df = weight_df.groupby(['trade_date', 'industryName1'])['weight'].sum().reset_index()
        pivot_df = pd.pivot_table(
            grouped_df, index='trade_date', columns='industryName1', values='weight').sort_index().fillna(0.)

        sector_df = pd.DataFrame(index=pivot_df.index, columns=industry_cluster_dict.keys())
        for key, value in industry_cluster_dict.items():
            value_include = [x for x in value if x in pivot_df.columns]
            sector_df[key] = pivot_df[value_include].sum(axis=1)

        return pivot_df, sector_df

    @staticmethod
    def _calculate_mkt_dis(portfolio_weight, security_value_df):
        df = pd.merge(
            portfolio_weight, security_value_df[['trade_date', 'ticker', 'MarketValue']],
            on=['trade_date', 'ticker']).dropna()
        df.loc[df['MarketValue'] < 100, 'sign'] = 'S'
        df.loc[(df['MarketValue'] >= 100) & (df['MarketValue'] < 300), 'sign'] = 'M'
        df.loc[(df['MarketValue'] >= 300) & (df['MarketValue'] < 1000), 'sign'] = 'L'
        df.loc[df['MarketValue'] >= 1000, 'sign'] = 'XL'
        grouped_df = df.groupby(['trade_date', 'sign'])['weight'].sum().reset_index()
        pivot_df = pd.pivot_table(
            grouped_df, index='trade_date', columns='sign', values='weight').sort_index().fillna(0.)
        pivot_df = pivot_df[['XL', 'L', 'M', 'S']].rename(
            columns={"XL": "1000亿以上", "L": "300-1000亿", "M": "100-300亿", "S": "100亿以下"})

        return pivot_df

    @staticmethod
    def plotly_area(df, title_text, range_upper=100, figsize=(1200, 600)):
        fig_width, fig_height = figsize
        cols = df.index.tolist()

        data = []
        for col in cols:
            tmp = go.Scatter(
                x=df.columns.tolist(),
                y=df.loc[col].values,
                name=col,
                mode='lines',
                line=dict(width=0.5),
                fill='tonexty',
                stackgroup='one')
            data.append(tmp)

        layout = go.Layout(
            title=title_text,
            autosize=False,
            width=fig_width,
            height=fig_height,
            showlegend=True,
            xaxis=dict(type='category'),
            yaxis=dict(
                type='linear',
                range=[1, range_upper],
                dtick=20,
                ticksuffix='%'))

        return data, layout

    @staticmethod
    def plotly_line(df, title_text, figsize=(1200, 600)):
        fig_width, fig_height = figsize
        data = []
        for col in df.columns:
            trace = go.Scatter(
                x=df.index.tolist(),
                y=df[col],
                name=col,
                mode="lines+markers"
            )
            data.append(trace)

        layout = go.Layout(
            title=dict(text=title_text),
            autosize=False, width=fig_width, height=fig_height,
            yaxis=dict(tickfont=dict(size=12), tickformat=',.0%', showgrid=True),
            xaxis=dict(showgrid=True),
            template='plotly_white'
        )

        return data, layout

    @staticmethod
    def plotly_double_y_line(df, title_text, figsize=(1200, 600)):
        fig_width, fig_height = figsize

        trace0 = go.Scatter(
            x=df.index.tolist(), y=df[df.columns[0]], mode="lines+markers", name=df.columns[0] + '(左轴)')
        trace1 = go.Scatter(
            x=df.index.tolist(), y=df[df.columns[1]], mode="lines+markers", name=df.columns[1] + '(右轴)', yaxis='y2')

        data = [trace0, trace1]

        layout = go.Layout(
            title=dict(text=title_text),
            autosize=False, width=fig_width, height=fig_height,
            yaxis=dict(tickfont=dict(size=12), showgrid=False),
            yaxis2=dict(overlaying='y', side='right'),
            xaxis=dict(showgrid=True),
            template='plotly_white'
        )

        return data, layout

    @staticmethod
    def draw_timeline_bar(df, title_text, min_value=0, min_height=600):
        num = int(df.gt(min_value).sum(axis=1).max())
        tl = Timeline(init_opts=opts.InitOpts(width='1200px', height='{}px'.format(max(25 * num, min_height))))

        for i in df.index.tolist():
            tmp = df.loc[i].dropna()
            tmp = tmp[tmp >= min_value].sort_values()
            bar = (
                Bar()
                .add_xaxis(tmp.index.tolist())
                .add_yaxis(
                    "持仓权重",
                    (tmp * 100).round(2).values.tolist(),
                    itemstyle_opts=opts.ItemStyleOpts(color="#37a2da"),
                    label_opts=opts.LabelOpts(is_show=False))
                .reversal_axis()
                .set_global_opts(
                    title_opts=opts.TitleOpts("{} (时间点: {})".format(title_text, i)),
                    xaxis_opts=opts.AxisOpts(
                        axislabel_opts=opts.LabelOpts(formatter="{value} %"),
                        splitline_opts=opts.SplitLineOpts(is_show=True, linestyle_opts=opts.LineStyleOpts(opacity=0.3)))
                )
            )
            tl.add(bar, "{}年{}月".format(i[:4], i[4:6]))

        return tl

    @staticmethod
    def plotly_table(portfolio_weight_df, threshold=1.0):
        df = portfolio_weight_df[portfolio_weight_df['ticker'].str[0].isin(['0', '3', '6', 'H'])]
        df['weight'] *= 100.
        df['content'] = df['sec_name'] + '（' + df['weight'].round(1).map(str) + '%）'
        df = df[df['weight'] >= threshold]
        max_len = int(df['trade_date'].value_counts().max())
        table_data = pd.DataFrame(index=np.arange(max_len), columns=sorted(df['trade_date'].unique()))
        for col in table_data.columns:
            tmp = df[df['trade_date'] == col].sort_values(by='weight', ascending=False)
            table_data.loc[:len(tmp) - 1, col] = tmp['content'].tolist()

        table_data.dropna(how='any', axis=0, inplace=True)

        fig = ff.create_table(table_data)

        return fig, table_data.shape[1]

    @staticmethod
    def plotly_change_table(portfolio_weight_df, threshold=1.0):
        df = portfolio_weight_df[portfolio_weight_df['ticker'].str[0].isin(['0', '3', '6', 'H'])]
        df['weight'] = (100 * df['weight']).round(1)
        pivot_df = pd.pivot_table(df, index=['ticker', 'sec_name'], columns='trade_date', values='weight').sort_index()
        pivot_df['mean'] = pivot_df.mean(axis=1)
        pivot_df = pivot_df[pivot_df.mean(axis=1) > threshold].sort_values(by='mean', ascending=False)
        del pivot_df['mean']
        pivot_df.fillna('-', inplace=True)
        res_df = pivot_df.reset_index().rename(columns={"ticker": "股票代码", "sec_name": "股票简称"})

        fig = ff.create_table(res_df)

        return fig, res_df.shape[1]

    def get_construct_result(self, is_tl_show=False):
        portfolio_weight = self.data_param['portfolio_weight']
        benchmark_weight = self.data_param['benchmark_weight']
        security_sector_df = self.data_param['security_sector_df']
        security_value_df = self.data_param['security_value_df']
        # 资产配置时序
        asset_allo_series = self._calculate_asset_allo_series(portfolio_weight)
        # 行业配置时序
        industry_allo_df, sector_allo_df = self._calculate_ind_allo_series(portfolio_weight, security_sector_df)
        industry_cr = \
            pd.concat([industry_allo_df.apply(lambda x: x.nlargest(1).sum(), axis=1).to_frame('第一大行业'),
                       industry_allo_df.apply(lambda x: x.nlargest(3).sum(), axis=1).to_frame('前三大行业'),
                       industry_allo_df.apply(lambda x: x.nlargest(5).sum(), axis=1).to_frame('前五大行业')], axis=1)
        # 重仓持股
        equity_weight = portfolio_weight[~portfolio_weight['ticker'].str[0].isin(['b', 'f', 'c'])]
        equity_weight = pd.pivot_table(
            equity_weight, index='trade_date', columns='sec_name', values='weight').sort_index()
        tmp = equity_weight.fillna(0.)
        equity_cr = pd.concat([tmp.apply(lambda x: x.nlargest(3).sum(), axis=1).to_frame('cr3'),
                               tmp.apply(lambda x: x.nlargest(5).sum(), axis=1).to_frame('cr5'),
                               tmp.apply(lambda x: x.nlargest(10).sum(), axis=1).to_frame('cr10')], axis=1)
        # 平均PE/PB
        value_df = pd.merge(security_value_df, portfolio_weight, on=['trade_date', 'ticker']).dropna(
            subset=['PB', 'PETTM'])
        # 做一下剔除
        value_df = value_df[(value_df['PETTM'] <= 1000) & (value_df['PETTM'] >= -1000)]
        average_pe = value_df.groupby('trade_date').apply(
            lambda x: (x['weight'] * x['PETTM']).sum() / x['weight'].sum()).to_frame('平均市盈率')
        average_pb = value_df.groupby('trade_date').apply(
            lambda x: (x['weight'] * x['PB']).sum() / x['weight'].sum()).to_frame('平均市净率')
        average_pe_pb = pd.concat([average_pe, average_pb], axis=1)
        # 持仓宽基分布
        df = portfolio_weight[portfolio_weight['ticker'].str[0].isin(['0', '3', '6'])]
        df = pd.merge(df, benchmark_weight, on=['trade_date', 'ticker'], how='left').fillna('other')
        bm_dis = df.groupby(['trade_date', 'benchmark_id'])['weight'].sum().reset_index()
        bm_dis = pd.pivot_table(
            bm_dis, index='trade_date', columns='benchmark_id', values='weight').sort_index().fillna(0.)
        map_dict = {"000300": "沪深300", "000905": "中证500", "000852": "中证1000", "other": "1800以外"}
        bm_dis.columns = [map_dict[x] for x in bm_dis.columns]
        # bm_dis = bm_dis[['沪深300', '中证500', '中证1000', '1800以外']]
        # 持仓市值分布
        mkt_dis = self._calculate_mkt_dis(portfolio_weight, security_value_df)

        upper_range = np.ceil(asset_allo_series.sum(axis=1).max() / 0.2) * 20
        data, layout = self.plotly_area(100 * asset_allo_series.T, '资产配置走势', upper_range)
        plot_render({"data": data, "layout": layout}, width=1200, height=600)

        upper_range = np.ceil(industry_allo_df.sum(axis=1).max() / 0.2) * 20
        data, layout = self.plotly_area(100 * industry_allo_df.T, '行业配置走势', upper_range)
        plot_render({"data": data, "layout": layout}, width=1200, height=600)
        # fig = go.Figure(data=data, layout=layout)
        # plot_ly(fig, filename="D:\\主观股多估值表基地\\图表所在\\B-行业配置走势.html", auto_open=False)

        upper_range = np.ceil(sector_allo_df.sum(axis=1).max() / 0.2) * 20
        data, layout = self.plotly_area(100 * sector_allo_df.T, '板块配置走势', upper_range)
        plot_render({"data": data, "layout": layout}, width=1200, height=600)

        data, layout = self.plotly_line(industry_cr, '行业集中度走势')
        plot_render({"data": data, "layout": layout}, width=1200, height=600)

        if is_tl_show:
            tl_bar = self.draw_timeline_bar(industry_allo_df, "截面行业权重", min_value=0.02, min_height=400)
            html_content = tl_bar.render_embed()
            print("%html {}".format(html_content))
            # tl_bar.render('D:\\主观股多估值表基地\\图表所在\\D-持仓行业明细.html')

        fig, n = self.plotly_table(portfolio_weight)
        plot_render({"data": fig.data, "layout": fig.layout}, width=120*n, height=600)

        fig, n = self.plotly_change_table(portfolio_weight)
        plot_render({"data": fig.data, "layout": fig.layout}, width=80*n, height=600)

        data, layout = self.plotly_line(equity_cr, "持股集中度走势")
        plot_render({"data": data, "layout": layout}, width=1200, height=600)

        if is_tl_show:
            tl_bar = self.draw_timeline_bar(equity_weight, "截面持股权重", min_value=0.02)
            html_content = tl_bar.render_embed()
            print("%html {}".format(html_content))

        data, layout = self.plotly_double_y_line(average_pe_pb.round(1), "持股估值水平走势")
        plot_render({"data": data, "layout": layout}, width=1200, height=600)

        upper_range = np.ceil(bm_dis.sum(axis=1).max() / 0.2) * 20
        data, layout = self.plotly_area(100 * bm_dis.T, '宽基成分配置走势', upper_range)
        plot_render({"data": data, "layout": layout}, width=1200, height=600)

        upper_range = np.ceil(mkt_dis.sum(axis=1).max() / 0.2) * 20
        data, layout = self.plotly_area(100 * mkt_dis.T, '持股市值分布走势', upper_range)
        plot_render({"data": data, "layout": layout}, width=1200, height=600)

class HoldingExtractor_DB:

    def __init__(self):

        self.total_asset_dmlist=['资产类合计:','资产合计']
        self.net_asset_dmlist=['基金资产净值:','资产净值','资产资产净值:']
        self.debt_dmlist=['负债类合计:']
        self.ashare_dmlist=['11020101','1102.01.01.','1101010101']
        self.szshare_dmlist=['11023101','1102.33.01.','1101013101']
        self.cyb_dmlistt=['11024101','1101014101']
        self.kcb_dmlist=['1102C101','1101C101']
        self.jjtz_dmlist = ['110104', '1105']

    @staticmethod
    def _shift_date(date):
        trade_dt = datetime.datetime.strptime(date, '%Y%m%d')
        pre_date = (trade_dt - datetime.timedelta(days=30)).strftime('%Y%m%d')

        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            pre_date, date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)

        trading_day_list = df[df['isOpen'] == 1]['calendarDate'].tolist()

        return trading_day_list[-1]

    @staticmethod
    def to_month_date(date_list):

        date_frame=pd.DataFrame(data=date_list,columns=['date'])
        date_frame['date']=date_frame['date'].astype(str)
        date_frame = date_frame[date_frame['date'] >= '20141031']
        date_frame['yeatmonth']=[x[0:6] for x in date_frame['date']]
        date_frame=date_frame.drop_duplicates('yeatmonth', keep='last')['date'].to_list()

        return  date_frame

    def _load_portfolio_weight(self,jjdm,date_list,fund_name=None):

        date_list=self.to_month_date(date_list)

        portfolio_weight_list = []
        for date in date_list:

            try:

                data=valuation.get_prod_valuation_by_jjdm_gzrq(jjdm,date)
                if(data.empty):
                    print("data for {0} missed,continue...".format(date))
                    continue
                date = self._shift_date(str(date))
                columns_map=dict(zip(['kmdm','sz','kmmc'],['科目代码','市值','科目名称']))

                data.rename(columns=columns_map,inplace=True)

                for dm in self.total_asset_dmlist:
                    if(len(data[data['科目代码'] == dm])>0):
                        total_asset = float(data[data['科目代码'] ==dm]['市值'].values[0])
                        break

                net_asset=0
                for dm in self.net_asset_dmlist:
                    if(len(data[data['科目代码'] == dm])>0):
                        net_asset = float(data[data['科目代码'] == dm]['市值'].values[0])
                        break

                if(net_asset==0):
                    for dm in self.debt_dmlist:
                        if (len(data[data['科目代码'] == dm]) > 0):
                            debt = float(data[data['科目代码'] == dm]['市值'].values[0])
                            net_asset=total_asset-debt
                            break

                leverage = total_asset / net_asset
                # A股
                sh=pd.DataFrame()
                for dm in self.ashare_dmlist:
                    if(len(data[data['科目代码'].str.startswith(dm)])>0):
                        sh = data[data['科目代码'].str.startswith(dm)]
                        break
                # sh = data[data['科目代码'].str.startswith('11020101')]
                sz=pd.DataFrame()
                for dm in self.szshare_dmlist:
                    if(len(data[data['科目代码'].str.startswith(dm)])>0):
                        sz = data[data['科目代码'].str.startswith(dm)]
                        break
                cyb = pd.DataFrame()
                for dm in self.cyb_dmlistt:
                    if (len(data[data['科目代码'].str.startswith(dm)]) > 0):
                        cyb = data[data['科目代码'].str.startswith(dm)]
                        break
                kcb = pd.DataFrame()
                for dm in self.kcb_dmlist:
                    if (len(data[data['科目代码'].str.startswith(dm)]) > 0):
                        kcb = data[data['科目代码'].str.startswith(dm)]
                        break

                if(len(sh)==0 and len(sz)==0 and len(cyb)==0 and len(kcb)==0):
                    print('No stock exist for {0} at {1}..'.format(jjdm,date))
                    continue

                equity_a = pd.concat([sh, sz, cyb, kcb], axis=0)

                equity_a['len'] = equity_a['科目代码'].apply(lambda x: len(x))
                equity_a = equity_a[equity_a['len'] > 10]
                if(len(equity_a)==0):
                    print('No stock details found in the data, date {} is skipped'.format(date))
                    continue

                if("." not in equity_a['科目代码'].values[0]):
                    equity_a['ticker'] = equity_a['科目代码'].apply(lambda x: x[-6:])
                else:
                    equity_a['ticker'] = equity_a['科目代码'].apply(lambda x: x[-8:-2])

                equity_a['weight'] = equity_a['市值'].astype(float) / net_asset
                equity_a = equity_a.rename(columns={"科目名称": "sec_name"})[['ticker', 'sec_name', 'weight']]
                # 港股
                hk1 = data[data['科目代码'].str.startswith('11028101')]
                hk2 = data[(data['科目代码'].str.startswith('11028201')) | (data['科目代码'].str.startswith('11028301'))]
                equity_hk = pd.concat([hk1, hk2], axis=0).fillna('')
                equity_hk['len'] = equity_hk['科目代码'].apply(lambda x: len(x))
                equity_hk = equity_hk[equity_hk['len'] > 8]
                equity_hk['ticker'] = equity_hk['科目代码'].apply(lambda x: x[-6:])
                equity_hk['科目名称'] = equity_hk['科目名称'].apply(lambda x: x.strip())  # add
                equity_hk['weight'] = equity_hk['市值'].astype(float) / net_asset
                equity_hk = equity_hk.rename(columns={"科目名称": "sec_name"})[['ticker', 'sec_name', 'weight']]
                equity_hk = equity_hk.groupby(['sec_name', 'ticker'])['weight'].sum().reset_index()
                # 债券
                tmp = data[data['科目代码'] == '1103']
                if tmp.empty:
                    bond_df = pd.DataFrame()
                else:
                    bond_ratio = tmp['市值'].astype(float).values[0] / net_asset
                    bond_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
                    bond_df.loc[0] = ['b00001', '债券投资', bond_ratio]
                # 基金
                for dm in self.jjtz_dmlist:
                    if (len(data[data['科目代码']==dm]) > 0):
                        tmp = data[data['科目代码'] == dm]
                        break
                if tmp.empty:
                    fund_df = pd.DataFrame()
                else:
                    fund_ratio = tmp['市值'].astype(float).values[0] / net_asset
                    fund_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
                    fund_df.loc[0] = ['f00001', '基金投资', fund_ratio]

                df = pd.concat([equity_a, equity_hk, bond_df, fund_df], axis=0)
                # 其他类
                cash_df = pd.DataFrame(columns=['ticker', 'sec_name', 'weight'])
                cash_df.loc[0] = ['c00001', '现金类投资', leverage - df['weight'].sum()]
                df = pd.concat([df, cash_df], axis=0)
                df['trade_date'] = date
                if(len(df[df['ticker'].str[0].isin(['0', '3', '6'])])>0):
                    portfolio_weight_list.append(df)

            except Exception as e:
                print(str(e)+"for jjdm:{0} at date :{1}".format(jjdm,date))
                continue

        if(len(portfolio_weight_list)>0):
            portfolio_weight_df = pd.concat(portfolio_weight_list)
        else:
            portfolio_weight_df = pd.concat([pd.DataFrame(columns=['ticker', 'sec_name','weight','trade_date'],
                                                          data=[['','',99999,date]])])

        portfolio_weight_df['jjdm'] = jjdm

        return  portfolio_weight_df



if __name__ == '__main__':
    name = '新方程泰暘臻选1期'
    # name = '新方程望正精英鹏辉'
    # name = '新方程域秀智享5号'
    # name = '亘曦2号'
    # name = '富乐一号'
    #name = '仁布积极进取1号'

    name='新方程巨杉,无股票'
    name= '同犇消费5号，20200630 资产合计项是空的'
    name='睿洪二号'
    name='丰岭稳健成长1期'
    name='望正精英-鹏辉1号'
    name='丹羿精选3号，无股票'
    name='亘曦2号,没有估值表'


    # jjdm = 'P02494'
    # # jjdm='P04365'
    # date_list =[ '20160630', '20160930', '20161230', '20170331', '20170630', '20170929']
    # # date_list = ['20150630', '20150930']
    # data=HoldingExtractor_DB()._load_portfolio_weight(jjdm=jjdm,date_list=date_list)
    # date_list=['20150630','20150930'])

    # name_list=['同犇1期']
    # code_list=['P04365']
    # date_list=pd.read_excel(r"E:\私募主观\私募主观可用持仓列表.xls")
    # for i in range(len(code_list)):
    #     code=code_list[i]
    #     name=name_list[i]
    #     date_list_temp=date_list[date_list['code']==code].astype(str)
    #     date_list_temp = date_list_temp[date_list_temp['date'] >= '20141031']
    #     date_list_temp['yeatmonth']=[x[0:6] for x in date_list_temp['date']]
    #     date_list_temp=date_list_temp.drop_duplicates('yeatmonth', keep='last')['date'].to_list()
    #     HoldingExtractor(table_name="subjective_fund_holding", fund_name=name,
    #                      is_increment=1, fund_code=code,date_list=date_list_temp)\
    #         .Pre_test(fromdb=True)
    #     print("{} test passed ".format(name))

    # name_list=['同犇1期']
    # code_list=['P04365']
    # date_list=pd.read_excel(r"E:\私募主观\私募主观可用持仓列表.xls")
    # for i in range(len(code_list)):
    #     code=code_list[i]
    #     name=name_list[i]
    #     date_list_temp=date_list[date_list['code']==code].astype(str)
    #     date_list_temp = date_list_temp[date_list_temp['date'] >= '20141031']
    #     date_list_temp['yeatmonth']=[x[0:6] for x in date_list_temp['date']]
    #     date_list_temp=date_list_temp.drop_duplicates('yeatmonth', keep='last')['date'].to_list()
    #     HoldingExtractor(table_name="subjective_fund_holding", fund_name=name, is_increment=1,fund_code=code,
    #                      date_list=date_list_temp)\
    #         .writeToDB(fromdb=True)
    #     print("{} done".format(name))

    HoldingAnalysor('SW5334', start_date='20150831', end_date='20210630').get_construct_result()


"""
套利和T0类策略收益统计
"""
from MyUtil.data_loader import get_trading_day_list, get_fund_nav_from_work, get_fund_nav_from_sql
import pandas as pd
import numpy as np


sql_params = {
    "ip": "192.168.223.152",
    "user": "readonly",
    "pass": "c24mg2e6",
    "port": "3306",
    "database": "work"
}

engine_params = "mysql+pymysql://{}:{}@{}:{}/{}".format(sql_params['user'], sql_params['pass'], sql_params['ip'],
                                                        sql_params['port'], sql_params['database'])


fund_list = ['SEF694', 'SNP300', 'SJB143', 'SE8723', 'SEK201', 'SGP290',
             'SJS027', 'SW0500',
             'SCP765', 'SLL241', 'SND168', 'SEP929', 'SGL202']

under_research = {
    "弘量君享量化CTA": "SQR403",
    "时代复兴微观一号": "SX1152",
    "弘源量化1号": "SK1274",
    "锐汇量化专享一号": "SGH587",
    "盛冠达时代匠心4号": "SJY896",
    "盛冠达股指套利3号": "SGS597",
    "博普量化对冲6号1期": "SLJ372"
}


def calculate(nav_series):
    ret_series = nav_series.pct_change().dropna()

    tmp = ret_series[ret_series.index >= '20210101']
    ret1 = (1 + tmp).prod() - 1

    tmp = ret_series[ret_series.index >= '20201119']
    ret2 = (1 + tmp).prod() - 1

    ret3 = np.power((1 + ret_series).prod(), (50. / len(ret_series))) - 1

    return [ret1, ret2, ret3]


if __name__ == '__main__':
    s_date = '20191119'
    e_date = '20211119'

    date_list = get_trading_day_list(s_date, e_date, frequency='week')

    data1 = []
    for f_id in fund_list:
        data1.append(get_fund_nav_from_work(s_date, e_date, f_id).set_index('t_date').reindex(date_list))

    data1 = pd.concat(data1)
    data1 = pd.pivot_table(data1.reset_index(), index='t_date', columns='name', values='nav').sort_index()
    data1.index.name = 'TRADEDATE'

    data2 = get_fund_nav_from_sql(s_date, e_date, under_research).reindex(date_list)

    nav_df = pd.concat([data1, data2], axis=1)


import numpy as np
import pandas as pd
from scipy import linalg
from cvxopt import solvers
solvers.options['show_progress'] = False
from cvxopt import matrix
from hbshare.fe.XZ import db_engine


class BL_Model:

    def __init__(self):


        self.hbdata=db_engine.HBDB()
        self.locdb=db_engine.PrvFunDB()

    def get_trade_date(self,end_date,fre,begin_date=None):

        #get the trading date for month or week end
        if(fre=='M'):
            fre_con="and SFYM='1'"
        elif(fre=='W'):
            fre_con="and SFZM='1'"
        else:
            raise Exception
            print('Please input correct frequency ')

        if(begin_date is None):
            begin_date_con=''
        else:
            begin_date_con=" and JYRQ>={0} ".format(begin_date)
        sql="SELECT JYRQ FROM funddb.JYRL where  JYRQ<={0} {1} {2}".format(end_date,fre_con,begin_date_con)
        trade_date=self.hbdata.db2df(sql,db='readonly').sort_values('JYRQ')['JYRQ'].values.tolist()
        return  trade_date

    def calculate_view_return(self,trade_month,asset_name):

        if(asset_name!='W00083'):
            sql="""select zqdm,jyrq,pe,roe,gxl from st_market.t_st_zs_hq where zqdm='{0}' and jyrq in ({1}) and jyrq>'20210930' """\
                .format(asset_name,','.join(trade_month))
            val_df1=self.hbdata.db2df(sql,db='alluser')

            #load daata from local db since NASDAQ roe data is missed from hbdb and no w000083 data is available in the hbdb
            sql="""select zqdm,jyrq,pe,roe,gxl from bl_valuation where zqdm='{0}' and jyrq in ({1}) and jyrq<='20210930'"""\
                .format(asset_name,','.join(trade_month))
            val_df2 =pd.read_sql(sql,con=db_engine.PrvFunDB().engine)
            val_df1=val_df1[val_df2.columns]
            val_df=pd.concat([val_df2,val_df1],axis=0).reset_index(drop=True)
        else:
            #load daata from local db since NASDAQ roe data is missed from hbdb and no w000083 data is available in the hbdb
            sql="""select zqdm,jyrq,pe,roe,gxl from bl_valuation where zqdm='{0}' and jyrq in ({1}) """\
                .format(asset_name,','.join(trade_month))
            val_df =pd.read_sql(sql,con=db_engine.PrvFunDB().engine)




        #calculate the PE median from the last 10 years(120 monthes)
        val_df['temp_pe']=val_df['pe'].rolling(120).median().shift(1)
        #calculate the implied return by temp_pe+roe+dividen_ratiio
        newpe=((val_df['temp_pe'][-1:]/val_df['pe'][-1:]).pow(1/5)-1).fillna(0)
        ret=newpe*100+val_df['gxl'][-1:]+val_df['roe'][-1:]
        ret.fillna(0,inplace=True)

        return  ret.values[0]/100

    def load_asset_return_data(self,trade_date,asset_name,asset_type):

        if (asset_type == 'index'):
            #read data from hbdb
            sql = "select rqzh,hb1z from st_market.t_st_zs_zhb where rqzh>= {0} and rqzh<={1} and zqdm='{2}' " \
                .format(trade_date[0], trade_date[-1], asset_name)

            ret_df = self.hbdata.db2df(sql, db='alluser')

            # #old saved return from local db with elder data that is missed in the hbdb
            # sql = "select rqzh,hb1z from bl_weekly_return where rqzh>= {0} and rqzh<={1} and zqdm='{2}' " \
            #     .format(trade_date[0], trade_date[-1], asset_name)
            # ret_df2 =pd.read_sql(sql,con=db_engine.PrvFunDB().engine)

            # ret_df=pd.concat([ret_df2,ret_df1],axis=0)
            ret_df=ret_df.sort_values('rqzh').reset_index(drop=True)
            ret_df.sort_values('rqzh')
            ret_df=ret_df[ret_df['hb1z'].abs()!=99999]


        elif (asset_type == 'public_fund'):

            sql="select  jzrq,fqdwjz from st_fund.t_st_gm_rhb where jjdm='{0}' and jzrq in ({1})"\
                .format(asset_name,','.join(trade_date))

            ret_df = self.hbdata.db2df(sql, db='funduser')
            ret_df['hb1z']=ret_df['fqdwjz'].pct_change()
            ret_df=ret_df[1:].drop('fqdwjz',axis=1)
            ret_df.rename(columns={'jzrq':'rqzh'},inplace=True)



        elif (asset_type == 'stock'):

            print('stock')
        else:
            raise Exception
            print('Please input correct asset_type, :index,stock,bond')


        ret_df['hb1z']=ret_df['hb1z']/100
        ret_df.rename(columns={'hb1z': 'ret_'+asset_name }, inplace=True)
        # ret_df.rename(columns={'hb1z': asset_name}, inplace=True)
        ret_df['rqzh'] = ret_df['rqzh'].astype(str)

        return  ret_df

    def cov_mat_from_return(self,trade_date,asset_list,asset_type):

        cov_mat=pd.DataFrame()
        cov_mat['Date']=trade_date

        for asset in asset_list:
            cov_mat=pd.merge(cov_mat,self.load_asset_return_data(trade_date,asset,asset_type),
                             how='inner',left_on='Date',right_on='rqzh').drop('rqzh',axis=1)


        cov_mat = cov_mat.drop('Date', axis=1).cov()*52

        return cov_mat

    def get_risk_aversion(self,end_date,engine):

        end_month=end_date[0:6]
        #get the monthly return of the bond return and calculate its mean as R_mkr:using 000012's return as market return temparary
        sql="select hb1y,tjyf from st_market.t_st_zs_yhb where zqdm='000012' and tjyf<='{0}'".format(end_month)
        deposit_rate_df=self.hbdata.db2df(sql,db='alluser')
        deposit_rate_df=deposit_rate_df[deposit_rate_df['hb1y']!=99999]
        deposit_rate_df['hb1y']=deposit_rate_df['hb1y']/100

        #get the monthly return of the market return and calculate its mean as R_mkr:using 000001's return as market return temparary
        sql="select hb1y,tjyf from st_market.t_st_zs_yhb where zqdm='000001'"
        mkt_ret_df=self.hbdata.db2df(sql,db='alluser')
        mkt_ret_df=mkt_ret_df[mkt_ret_df['hb1y']!=99999]
        mkt_ret_df['hb1y']=mkt_ret_df['hb1y']/100


        risk_df=pd.merge(mkt_ret_df,deposit_rate_df,how='inner',left_on='tjyf',right_on='tjyf')
        rf=risk_df['hb1y_y'].mean()
        rmkr=risk_df['hb1y_x'].mean()
        varmkr=risk_df['hb1y_x'].var()
        risk_aversion=(rmkr-rf)/varmkr


        #save the risk aversion into local db
        inputdf=pd.DataFrame()
        inputdf['risk_aversion']=[risk_aversion]
        inputdf['date']=end_month

        #check if data already exist
        sql="select count(*) c from bl_risk_aversion where date='{0}' "\
            .format(end_month)
        if(pd.read_sql(sql,con=engine)['c'].values[0]>0):
            sql = "delete from bl_risk_aversion where date='{0}'"\
                .format(end_month)
            engine.execute(sql)

        inputdf.to_sql('bl_risk_aversion',con=engine,index=False,if_exists='append')

        return risk_aversion

    def get_weight(self,end_date,asset_list,engine):

        weightdf=pd.DataFrame(data=asset_list,columns=['name'])

        sql="select code,mkr_value from bl_market_value where Date='{0}' and code in ({1})"\
            .format(end_date,"'"+"','".join(asset_list)+"'")
        weightdf=pd.merge(weightdf,pd.read_sql(sql,con=engine),left_on='name',right_on='code',how='left')
        weight_list=weightdf['mkr_value'].values.tolist()

        # sql="""select zqdm,zsz from st_market.t_st_zs_hq where zqdm in ({0}) and jyrq = '{1}' """\
        #     .format("'"+"','".join(asset_list)+"'",end_date)
        # val_df=self.hbdata.db2df(sql,db='alluser')
        # weightdf=pd.merge(weightdf,self.hbdata.db2df(sql,db='alluser'),left_on='name',right_on='zqdm',how='left')
        # weight_list=weightdf['mkr_value'].values.tolist()

        return weight_list

    def implied_return_from_weight(self,mkt_val,risk_aversion,asset_list,cov_matrix):

        weights=np.array(mkt_val)/sum(mkt_val)
        implied_ret=risk_aversion*np.dot(cov_matrix,weights)

        return implied_ret

    def set_boundary(self,asset_num,lb,ub):

        G = matrix(np.vstack((np.diag([-1]*asset_num), np.diag([1]*asset_num))), tc='d')
        h = matrix(np.array(lb+ub), tc='d') # 为各参数的上下限！！！！
        A = matrix(np.array([[1]*asset_num]), tc='d')
        b = matrix(np.array([1]), tc='d')

        boundary=dict()
        boundary['G']=G
        boundary['h'] = h
        boundary['A'] = A
        boundary['b'] = b

        return boundary

    def cov2db(self,cov_matrix,engine,end_date,version,table_sur_name):

        # write cov_matrix into database
        inputdata=cov_matrix.copy()
        inputdata['row_name'] = inputdata.index
        inputdata['Date'] = end_date
        table_name=table_sur_name+version

        #check if table exist already
        sql="SELECT count(*) as c  FROM information_schema.tables WHERE table_name = '{0}'"\
            .format(table_name)
        if(pd.read_sql(sql,con=engine)['c'].values[0]>0):

            sql="select count(*) as c from {0} where Date='{1}' ".format(table_name,end_date)
            # check if data already exist,if yes delete the old data and replace with new one
            if(pd.read_sql(sql,con=engine)['c'].values[0]>0):
                sql="delete from {0} where Date={1}".format(table_name,end_date)
                engine.execute(sql)

        inputdata.to_sql(table_name, engine, index=False, if_exists='append')

    def return2db(self,tabel_name,ret_list,engine,asset_list,end_date,version,col):

        # write implied return into database
        inputdata=pd.DataFrame()
        inputdata['code']=asset_list
        inputdata[col] = ret_list
        inputdata['Date'] = end_date
        inputdata['version']=version

        #check if data already exist
        sql="select count(*) c from {2} where version='{0}' and Date='{1}' and code in ({3})"\
            .format(version,end_date,tabel_name,"'"+"','".join(asset_list)+"'")
        if(pd.read_sql(sql,con=engine)['c'].values[0]>0):
            sql = "delete from {2} where Date='{0}' and version= '{1}'"\
                .format( end_date,version,tabel_name)
            engine.execute(sql)

        inputdata.to_sql('{0}'.format(tabel_name), engine, index=False, if_exists='append')

    def mkrvalue2db(self,asset_list,mkr_v,end_date,engine):

        # write implied return into database
        inputdata=pd.DataFrame()
        inputdata['code']=asset_list
        inputdata['mkr_value'] = mkr_v
        inputdata['Date'] = end_date

        #check if data already exist
        sql="select distinct code from bl_market_value where Date='{0}'".format(end_date)
        current_list=pd.read_sql(sql,con=engine)['code'].values
        delete_list=list(set(current_list)&set(asset_list))
        if(len(delete_list)>0):
            sql="delete from bl_market_value where Date='{0}' and code in ({1})"\
                .format(end_date,"'"+"','".join(delete_list)+"'")
            engine.execute(sql)

        inputdata.to_sql('bl_market_value', engine, index=False, if_exists='append')

    def read_cov_fromdb(self,engine,asset_list,version,end_date,asset_type):

        table_name='bl_'+asset_type+'_cov_'+version
        #get the cov matrix from local database
        sql="select {0},row_name from {1} where Date='{2}' "\
            .format('ret_'+',ret_'.join(asset_list),table_name,end_date)

        cov_matrix=pd.read_sql(sql,con=engine)
        dfT=cov_matrix.T
        dfT.columns=cov_matrix['row_name']
        dfT=dfT[['ret_'+str(x) for x in  asset_list]]
        cov_matrix=dfT.T.drop('row_name',axis=1)
        #make sure the cov_matrix order is consistent with asset list
        cov_matrix=cov_matrix[['ret_'+str(x) for x in  asset_list]]

        return cov_matrix.astype(float)

    def read_return_fromdb(self,asset_list,engine,version,end_date,table_name,col):

        sql="select code,{4} from {3} where Date='{0}' and version='{1}' and code in ({2})"\
            .format(end_date,version,"'"+"','".join(asset_list)+"'",table_name,col)
        asset_order=pd.DataFrame()
        asset_order['Code']=asset_list
        tempdf=pd.merge(asset_order,pd.read_sql(sql,con=engine),how='left',left_on='Code',right_on='code')

        return tempdf[col].values

    def update_asset_pool(self,dir,version):

        engine = db_engine.PrvFunDB().engine
        #delete data if already exists
        sql="delete from bl_assets_pool where version='{}'".format(version)
        engine.execute(sql)

        local_file=pd.read_excel(dir)
        local_file['version']=version
        local_file.to_sql('bl_assets_pool',con=engine,index=False, if_exists='append')
        print('Update asset pool table Done..')

    def update_valuation_data(self):

        local_file=pd.read_excel(r"E:\GitFolder\hbshare\fe\AAM\blmodel\PE_ROE_dividend.xlsx")
        local_file.to_sql('bl_valuation',con=self.locdb.engine,index=False,if_exists='append')
        print('valuation data updated successfully')

    def update_marketvalue_data(self):

        local_file=pd.read_excel(r"E:\GitFolder\hbshare\fe\AAM\blmodel\monthly MV.xlsx")
        local_file.to_sql('bl_market_value',con=self.locdb.engine,index=False,if_exists='append')
        print('marketvalue_data updated successfully')

    def blm_solver(self,sigma,mu,P,Q,Omega,delta,constrains):

        prio_weight=linalg.inv(sigma)
        med_weight=np.dot(np.dot(P,linalg.inv(Omega)),P.T)

        post_cov = linalg.inv(prio_weight + med_weight)

        post_ret=np.dot(post_cov,(np.dot(mu,prio_weight)+np.dot(np.dot(P,linalg.inv(Omega)),np.array(Q).T)))

        # M = np.dot(np.dot(P, sigma), P.T) + Omega
        # M = M.astype(float)
        # M = linalg.inv(M)
        # er = mu.T + np.dot(np.dot(np.dot(sigma, P.T), M), (Q - np.dot(P, mu.T)))

        # 插入限制条件正向求解马克维茨方程
        p = matrix(delta * post_cov, tc='d')
        q = matrix(-post_ret, tc='d')

        G=constrains['G']
        h = constrains['h']
        A = constrains['A']
        b = constrains['b']
        sol = solvers.qp(p, q, G, h, A, b)

        return [ x for x in sol['x']]

    def bl_model_index_data_preparation(self, end_date, asset_list,version):

        engine=db_engine.PrvFunDB().engine

        #########################################for index asset #########################################
        # for index asset
        # get last 10 years trade month date
        trade_date_month = self.get_trade_date(end_date, fre='M')[-200:]

        # last 6 years weekly data is needed for calculating the cov martix
        begin_date = str(int(end_date[0:4]) - 6) + end_date[4:6] + '01'
        trade_date_week = self.get_trade_date(end_date=end_date, fre='W', begin_date=begin_date)

        # calcualte the Cov matrix based on weekly return for the laset 6years
        cov_matrix = self.cov_mat_from_return(trade_date_week, asset_list, asset_type='index')

        # write cov_matrix into database
        self.cov2db(cov_matrix, engine, end_date, version,'bl_index_cov_')

        # calcualte the risk_aversion
        risk_aversion = self.get_risk_aversion(end_date,engine)

        # calculate the prio return based on the market value weights, COV matrix and risk aversion
        mkt_val=self.get_weight(trade_date_month[-1],asset_list,engine)

        prio_return = self.implied_return_from_weight(mkt_val=mkt_val, risk_aversion=risk_aversion,
                                                     asset_list=asset_list, cov_matrix=cov_matrix)
        # write implied return into database
        self.return2db('bl_implied_return',prio_return, engine, asset_list, end_date, version,'implied_ret')
        #self.implied_ret2db(prio_return, engine, asset_list, end_date, version)


        # calculate the view return if necessary
        # use the last 10 years data for calculating view return
        view_ret = []
        for asset in asset_list:
            view_ret.append(self.calculate_view_return(trade_date_month, asset))

        # write view return into database
        #self.viewret2db(view_ret, engine, asset_list, end_date, version)
        self.return2db('bl_view_return', view_ret, engine, asset_list, end_date, version,'view_ret')

    def bl_model_publicfund_data_preparation(self,end_date,asset_list,version):

        engine = db_engine.PrvFunDB().engine

        #########################################for public fund asset #########################################
        # for index asset

        #last 3.5 years/ 182 weeks  weekly data is needed for calculating the cov martix
        trade_date_week=self.get_trade_date(end_date=end_date,fre='W')[-(182+1):]

        #calcualte the Cov matrix based on weekly return for the laset  3.5 years/ 182 weeks
        cov_matrix=self.cov_mat_from_return(trade_date_week,asset_list,asset_type='public_fund')

        # write cov_matrix into database
        self.cov2db(cov_matrix, engine, end_date, version,'bl_public_fund_cov_')

        #calcualte the risk_aversion
        risk_aversion=self.get_risk_aversion(end_date,engine)

        #calculate the prio return based on the market value weights, COV matrix and risk aversion
        mkt_val = [100]*len(asset_list)

        prio_return=self.implied_return_from_weight(mkt_val=mkt_val,risk_aversion=risk_aversion,
            asset_list=asset_list,cov_matrix=cov_matrix)

        # write implied return into database
        self.return2db('bl_implied_return',prio_return, engine, asset_list, end_date, version,'implied_ret')


        view_ret = []
        trade_date_week=trade_date_week[-(105+1):]
        for asset in asset_list:
            view_ret.append(self.load_asset_return_data(trade_date_week,asset,'public_fund')['ret_'+asset].mean()*52)

        # write view return into database
        self.return2db('bl_view_return', view_ret, engine, asset_list, end_date, version,'view_ret')



if __name__ == '__main__':


        end_date='20220131'
        version='20220131_v1'
        blm=BL_Model()
        localdb_engine=db_engine.PrvFunDB().engine

        #blm.update_valuation_data()
        #blm.update_asset_pool(r"E:\GitFolder\hbshare\fe\AAM\blmodel\asset_pool_update.xls", version)
        #blm.update_marketvalue_data()


        # updates the market value table if necessary
        # mkt_val=[]
        # asset_list=[]
        # blm.mkrvalue2db(asset_list,mkt_val,'20210827',db_engine.PrvFunDB().engine)

        #get the assets from pool
        sql=" select * from bl_assets_pool where version='{0}' ".format(version)
        assets_df=pd.read_sql(sql, con=localdb_engine)
        index_assets=assets_df[assets_df['asset_type']=='index']['code'].values.tolist()
        public_funds = assets_df[assets_df['asset_type'] == 'public_fund']['code'].values.tolist()

        blm.bl_model_index_data_preparation(end_date=end_date,asset_list=index_assets,version=version)
        blm.bl_model_publicfund_data_preparation(end_date=end_date, asset_list=public_funds,version=version)


"""
套利FOF回测模块
"""
import numpy as np
from MyUtil.data_loader import get_trading_day_list, get_fund_nav_from_sql
from MyUtil.util_func import cal_annual_return, cal_annual_volatility, cal_max_drawdown, cal_sharpe_ratio
import pandas as pd
import datetime


if __name__ == '__main__':
    s_date = '20200103'
    e_date = '20211126'

    f_dict = {
        "展弘稳进1号1期": "SE8723",
        "盛冠达股指套利3号": "SGS597",
        "稳博中睿6号": "SJB143"
    }

    date_list = get_trading_day_list(s_date, e_date, frequency='week')

    nav_df = get_fund_nav_from_sql(s_date, e_date, f_dict).reindex(date_list).fillna(method='ffill')

    mx_df = pd.read_excel('D:\\蒙玺竞起6号净值.xlsx', sheet_name=0)
    mx_df['trade_date'] = mx_df['trade_date'].apply(lambda x: datetime.datetime.strftime(x, '%Y%m%d'))
    mx_df.rename(columns={"nav": "蒙玺竞起6号"}, inplace=True)
    mx_df = mx_df.set_index('trade_date').reindex(date_list).dropna()

    nav_df = pd.concat([nav_df, mx_df], axis=1)
    return_df = nav_df.pct_change().fillna(0.)

    # 等权
    weight_df1 = pd.DataFrame(index=return_df.index, columns=return_df.columns)
    weight_df1.loc['20200103', :] = [1/3, 1/3, 1/3, 0]

    rbl_list = ['20201120']

    for i in range(1, len(return_df)):
        date = return_df.index[i]
        weight_df1.iloc[i, :] = weight_df1.iloc[i - 1, :] * (1 + return_df.iloc[i, :])
        if date in rbl_list:
            weight_df1.loc[date] = weight_df1.loc[date].mean()

    port_nav1 = weight_df1.sum(axis=1)

    # 额度加权
    weight_df2 = pd.DataFrame(index=return_df.index, columns=return_df.columns)
    weight_df2.loc['20200103', :] = [4/9, 1/9, 4/9, 0]
    weight_df2.loc['20201120', :] = [4/11, 1/11, 4/11, 2/11]

    rbl_list = ['20201120']

    for i in range(1, len(return_df)):
        date = return_df.index[i]
        weight_df2.iloc[i, :] = weight_df2.iloc[i - 1, :] * (1 + return_df.iloc[i, :])
        if date in rbl_list:
            weight_df2.loc[date] = [weight_df2.loc[date].sum() * x for x in [4/11, 1/11, 4/11, 2/11]]

    port_nav2 = weight_df2.sum(axis=1)

    # 净值
    port_nav = pd.concat([port_nav1.to_frame('等权组合'), port_nav2.to_frame('最大额度组合')], axis=1)
    # 累计收益率
    cum_return = port_nav.iloc[-1] - 1
    # 年化收益
    an_return = port_nav.pct_change().dropna(how='all').apply(cal_annual_return, axis=0)
    # 年化波动
    an_vol = port_nav.pct_change().dropna(how='all').apply(cal_annual_volatility, axis=0)
    # 最大回撤
    max_draw = port_nav.apply(cal_max_drawdown, axis=0)
    draw_series = port_nav.apply(lambda x: x / x.cummax() - 1)
    # sharpe
    sharpe = port_nav.pct_change().dropna(how='all').apply(lambda x: cal_sharpe_ratio(x, 0.015), axis=0)
    # # sortino
    down_std = port_nav.pct_change().dropna(how='all').apply(lambda x: x[x < 0].std() * np.sqrt(52), axis=0)
    sortino = an_return / down_std
    # calmar
    calmar = an_return / max_draw.abs()
    # 胜率
    win_ratio = port_nav.pct_change().dropna(how='all').apply(lambda x: x.gt(0).sum() / len(x), axis=0)
    # 平均损益比
    win_lose = port_nav.pct_change().dropna(how='all').apply(lambda x: x[x > 0].mean() / x[x < 0].abs().mean(), axis=0)
    # 年度累计
    ret_20 = port_nav.loc['20201231'] - 1
    ret_21 = port_nav.iloc[-1] / port_nav.loc['20201231'] - 1


import datetime

import pymysql
import pandas as pd
import requests as rq
from typing import List, Union, Optional, Dict, Tuple, Any
from collections import defaultdict
from .database_connection import *
from .database_dictionary import *


def sql_request(
		sql: str, db: str = "readonly"
) -> List[Dict]:
	init_hbs()
	post_data = []
	for i in range(1, int(1e10)):
		data = hbs.db_data_query(db, sql, is_pagination=True, page_num=i, page_size=5000)
		if data.get("success", False):
			now_data = data.get("data", [])
			if len(now_data) == 0:
				break
			post_data.extend(now_data)
		else:
			raise Exception(f'请求错误： {data.get("returnMsg", "无详细错误信息")}')
	return post_data


def base_request(
		fields: Union[None, str, List[str]],
		conditions: Union[List[str], None],
		table: str,
) -> List[dict]:
	init_hbs()
	if not fields:
		fields = "*"
	if isinstance(fields, list):
		fields = ",".join(fields)
	if isinstance(conditions, list):
		sql = f"select {fields} from {table} where {','.join(conditions)}"
	else:
		sql = f"select {fields} from {table}"
	post_data = sql_request(sql)
	return post_data


def get_manager_basic_info(
		fields: Union[None, str, List[str]],
		conditions: Union[List[str], None]
) -> pd.DataFrame:
	data = base_request(fields, conditions, "funddb.jjgg")
	data_df = pd.DataFrame(data).drop(["ROW_ID"], axis=1)
	data_df.columns = [ManagerDict.get(x.upper(), x) for x in data_df.columns]
	return data_df


def get_public_fund_net_value(
		fields: Union[None, str, List[str]],
		conditions: Union[None, List[str]]
) -> pd.DataFrame:
	if conditions is None:
		conditions = []
	conditions.append("jjdm in (select jjdm from funddb.jjxx1)")
	data = base_request(fields, conditions, "funddb.jjjz")
	data_df = pd.DataFrame(data).drop(["ROW_ID"], axis=1)
	data_df.columns = [FundNetValueDict.get(x.upper(), x) for x in data_df.columns]
	return data_df


def get_all_fund_category(is_first: bool = True) -> Dict[str, str]:
	if is_first:
		type_code = "301634"
	else:
		type_code = "301635"
	data = sql_request(f"SELECT typecode, typename FROM st_main.t_st_gg_mj where  type='{type_code}' order by typecode", "funduser")
	result = dict()
	for da in data:
		result[da['typecode']] = da['typename']
	return result


def get_csi_share_index(start_date: Optional[datetime.datetime] = None) -> List[dict]:
	start_date = '20060101' if start_date is None else start_date
	all_share_sql = f"select jyrq, spjg from funddb.zsjy where ZQDM='000985' and spjg is not null and jyrq >= {start_date} order by jyrq"
	return sql_request(all_share_sql)


def get_csi_bond_index(start_date: Optional[datetime.datetime] = None) -> List[dict]:
	start_date = '20060101' if start_date is None else start_date
	all_bond_sql = f"select jyrq, spjg from funddb.zsjy where ZQDM='H11001' and spjg is not null and jyrq >= {start_date} order by jyrq"
	return sql_request(all_bond_sql)


def get_fund_category(funds: List[str]) -> List[dict]:
	fund_category_sql = f"""
			SELECT a.jjdm, b.typename as first, c.typename as second
			From st_fund.t_st_gm_jjxx a
			join st_main.t_st_gg_mj b on a.jjfl = b.typecode
			join st_main.t_st_gg_mj c on a.ejfl = c.typecode
			where b.type='301634' and c.type = '301635' and a.jjdm in ('{"','".join(funds)}') and a.jjzt<>'c'
	"""
	return sql_request(fund_category_sql, "funduser")


def get_net_worth_by_manager(manager_code: str, start_date: Optional[str] = None) -> List[dict]:
	date = '20060101' if start_date is None else start_date
	sql = f"""
		select a.rydm,a.ryxm,a.rzrq,case when a.lrrq='19000101' then null else lrrq end as lrrq,a.jjdm,b.jjjc, b.clrq, c.jzrq,c.jjjz,c.ljjz 
		from funddb.jjgg a
		join funddb.jjxx1 b on a.jjdm=b.jjdm 
		join funddb.jjjz c on a.jjdm=c.jjdm and c.jzrq>=a.rzrq and c.jzrq>='{date}' and c.jzrq<=case when a.lrrq='19000101' or a.lrrq is null then to_char(sysdate,'yyyymmdd') else lrrq  end
		where a.rydm='{manager_code}'  
		order by c.jzrq, a.jjdm DESC
		"""
	return sql_request(sql, "readonly")


def get_fund_position_ratio(fund_code: str) -> List[dict]:
	sql = f"""
	select gpbl, zqbl, jjdm, jsrq 
	from funddb.zcpz
	where  ggrq is not null  and jjdm = '{fund_code}'
	order by ggrq
	"""
	return sql_request(sql, "readonly")


def get_fund_size(funds: List[str]) -> List[dict]:
	sql = "select jzzc, jjdm, ggrq from funddb.zcpz where jjdm in ('" + "','".join(map(str, funds)) + "') order by ggrq, jjdm"
	return sql_request(sql, "readonly")


def get_all_trading_day(start_date: Union[str, None] = None) -> List[dict]:
	start_date = '20060101' if start_date is None else start_date
	now_date = datetime.datetime.now().date().strftime("%Y%m%d")
	sql = f"select jyrq from funddb.jyrl where sfjj = 0 and jyrq > '{start_date}' and jyrq < '{now_date}' order by jyrq"
	return sql_request(sql, "readonly")


def get_fund_net_worth_days(start_date: Union[str, None] = None) -> List[dict]:
	start_date = '20060101' if start_date is None else start_date
	sql = f"""
		select a.jjdm, count(a.jzrq) as count, max(a.jzrq) as max 
		from funddb.jjjz a 
		left join funddb.jjxx1 b on a.jjdm = b.jjdm 
		where a.jjjz is not null and a.jzrq > '{start_date}' and b.cpfl = '2' and b.jjfl in ('1', '2', '3') 
		group by a.jjdm
	"""
	return sql_request(sql, "readonly")


def get_latest_mix_fund_announcements_by_manager(manager_code: str) -> List[dict]:
	sql = f"""
		select a.jjdm, max(b.jsrq) as latest_date
		from funddb.jjgg a
		join funddb.zcpz b on a.jjdm = b.jjdm
		join funddb.jjxx1 c on a.jjdm = c.jjdm
		where a.rydm = '{manager_code}' and  c.jjfl = '3' and b.jsrq < case when a.lrrq='19000101' or a.lrrq is null then to_char(sysdate,'yyyymmdd') else a.lrrq  end
		group by a.jjdm
	"""
	return sql_request(sql, "readonly")


def insert_net_worth_record(
		last_record_time: datetime,
		manager_code: str,
		index_value: float,
		is_share: bool = True,
		mix_fund_latest_announcement: Optional[dict] = None
):
	now_time = datetime.datetime.now()
	record_type = 'share' if is_share else 'bond'
	NetWorthRecord.create(
		index_date=last_record_time,
		fund_manager_code=manager_code,
		record_date=now_time,
		record_type=record_type,
		index_value=index_value,
		mix_fund_latest_announcement=str(mix_fund_latest_announcement),
	)


def get_latest_net_worth_record(fund_manager_code: str) -> List[dict]:
	import datetime
	result = []
	for record_type in ['share', 'bond']:
		try:
			record = (
				NetWorthRecord
				.select(
					fn.MAX(NetWorthRecord.record_date),
					NetWorthRecord.index_date,
					NetWorthRecord.record_type,
					NetWorthRecord.index_value,
					NetWorthRecord.mix_fund_latest_announcement
				)
				.where(NetWorthRecord.fund_manager_code == fund_manager_code, NetWorthRecord.record_type == record_type)
				.group_by(NetWorthRecord.fund_manager_code)
				.order_by(-NetWorthRecord.record_date)
				.get()
			)
			record = dict(
				record_date=record.record_date,
				index_date=record.index_date,
				record_type=record_type,
				index_value=record.index_value,
				mix_fund_latest_announcement=eval(record.mix_fund_latest_announcement)
			)
		except:
			record = dict()
		result.append(record)
	return result


def get_share_index_by_manager(manager_code: str):
	history_records = (
		ShareIndex
		.select()
		.where(ShareIndex.manager_code == manager_code)
		.order_by(ShareIndex.index_date)
	)
	try:
		return pd.DataFrame(list(history_records.to_dicts()))
	except:
		return pd.DataFrame()


def get_bond_index_by_manager(manager_code: str):
	history_records = (
		BondIndex
		.select()
		.where(BondIndex.manager_code == manager_code)
		.order_by(BondIndex.index_date)
	)
	try:
		return pd.DataFrame(list(history_records.to_dicts()))
	except:
		return pd.DataFrame()


def save_share_index_values(manager_code: str, index_values: List[dict]):
	to_be_deleted = (
		ShareIndex
		.delete()
		.where(ShareIndex.manager_code == manager_code, ShareIndex.index_date.in_([x['date'] for x in index_values]))
	)
	try:
		to_be_deleted.execute()
	except:
		pass
	index = [{'manager_code': manager_code, 'index_date': x['date'], 'index_value': x['index']} for x in index_values]
	ShareIndex.insert_many(index).execute()


def save_bond_index_values(manager_code: str, index_values: List[dict]):
	to_be_deleted = (
		BondIndex
		.delete()
		.where(BondIndex.manager_code == manager_code, BondIndex.index_date.in_([x['date'] for x in index_values]))
	)
	try:
		to_be_deleted.execute()
	except:
		pass
	index = [{'manager_code': manager_code, 'index_date': x['date'], 'index_value': x['index']} for x in index_values]
	BondIndex.insert_many(index).execute()


def create_tables():
	record_db.create_tables([NetWorthRecord, ShareIndex, BondIndex])



import numpy as np
import pandas as pd
from scipy.stats import norm
import matplotlib.pyplot as plt
from datetime import datetime


def bsm_d1(K, F, sigma, r, T, q):
    d1 = (np.log(F/K) + (r - q + sigma**2/2) * T)/(sigma * T**0.5)
    return d1


def bsm(K, F, sigma, r, T, q, direction):
    # 欧式 BSM
    d1 = bsm_d1(K, F, sigma, r, T, q)
    d2 = d1 - sigma * (T**0.5)
    if direction.lower() == 'call':
        c = np.exp(-q * T) * F * norm.cdf(d1) - np.exp(-r * T) * K * norm.cdf(d2)
        return c
    elif direction.lower() == 'put':
        p = np.exp(-r * T) * K * norm.cdf(-d2) - np.exp(-q * T) * F * norm.cdf(-d1)
        return p
    elif direction.lower() == 'binary call':
        bc = np.exp(-r * T) * norm.cdf(d2)
        return bc
    elif direction.lower() == 'binary put':
        bp = np.exp(-r * T) * norm.cdf(-d2)
        return bp
    else:
        raise ValueError("Parameter direction Error")


def bsm_delta(K, F, sigma, r, T, q, direction):
    # 欧式 delta
    d1 = bsm_d1(K, F, sigma, r, T, q)
    d2 = d1 - sigma * (T ** 0.5)
    if direction.lower() == 'call':
        return np.exp(-q * T) * norm.cdf(d1)
    elif direction.lower() == 'put':
        return np.exp(-q * T) * (norm.cdf(d1) - 1)
    elif direction.lower() == 'binary call':
        return np.exp(-r * T) * norm.pdf(d2) / (sigma * F * T**0.5)
    elif direction.lower() == 'binary put':
        return -np.exp(-r * T) * norm.pdf(d2) / (sigma * F * T**0.5)
    else:
        raise ValueError("parameter direction should either be 'call' or 'put'")


def bsm_gamma(K, F, sigma, r, T, q):
    d1 = bsm_d1(K, F, sigma, r, T, q)
    gamma = norm.pdf(d1) * np.exp(-q * T) / (F * sigma * T**0.5)
    return gamma


def bsm_theta(K, F, sigma, r, T, q, direction):
    d1 = bsm_d1(K, F, sigma, r, T, q)
    d2 = d1 - sigma * (T ** 0.5)
    if direction.lower() == 'call':
        theta = -F * norm.pdf(d1) * sigma * np.exp(-q * T) / (2 * T**0.5) \
                + q * F * norm.cdf(d1) * np.exp(-q * T) - r * K * np.exp(-r * T) * norm.cdf(d2)
        return theta
    elif direction.lower() == 'put':
        theta = -F * norm.pdf(d1) * sigma * np.exp(-q * T) / (2 * T ** 0.5) \
                - q * F * norm.cdf(-d1) * np.exp(-q * T) + r * K * np.exp(-r * T) * norm.cdf(-d2)
        return theta
    else:
        raise ValueError("parameter direction should either be 'call' or 'put'")


def bsm_vega(K, F, sigma, r, T, q):
    d1 = bsm_d1(K, F, sigma, r, T, q)
    vega = F * T**0.5 * norm.pdf(d1) * np.exp(-q * T)
    return vega


def kv(K, F, sigma, r, T, q, direction):
    # 亚式看涨 KV
    q = q + sigma ** 2 / 6
    sigma = sigma / np.sqrt(3)
    d1 = bsm_d1(K, F, sigma, r, T, q)
    d2 = d1 - sigma * (T**0.5)
    if direction.lower() == 'call':
        c = np.exp(-q * T) * F * norm.cdf(d1) - np.exp(-r * T) * K * norm.cdf(d2)
        return c
    elif direction.lower() == 'put':
        p = np.exp(-r * T) * K * norm.cdf(-d2) - np.exp(-q * T) * F * norm.cdf(-d1)
        return p
    else:
        raise ValueError("parameter direction should either be 'call' or 'put'")


def baw(K, F, sigma, r, T, q, direction, epsilon):
    n = 2 * (r - q) / (sigma**2)
    m = 2 * r / (sigma**2)
    k = 1 - np.exp(-r * T)
    q1 = (1 - n - np.sqrt((n - 1)**2 + 4 * m / k))/2
    q2 = (1 - n + np.sqrt((n - 1)**2 + 4 * m / k))/2

    dx = 0.01
    # epsilon = 0.00001
    i = 1
    imax = 50

    if direction.lower() == 'call':
        if q == 0:
            Call = bsm(K, F, sigma, r, T, q, 'call')
        else:
            S1 = K
            while i <= imax:
                c0 = bsm(K, S1, sigma, r, T, q, 'call')
                cu = bsm(K, S1 * (1 + dx), sigma, r, T, q, 'call')
                cd = bsm(K, S1 * (1 - dx), sigma, r, T, q, 'call')
                gap0 = S1 - K - c0 - (S1 / q2) * (1 - np.exp(-q * T) * norm.cdf(bsm_d1(K, S1, sigma, r, T, q)))
                gapu = S1 * (1 + dx) - K - cu - (S1 * (1 + dx) / q2) * (
                        1 - np.exp(-q * T) * norm.cdf(bsm_d1(K, S1 * (1 + dx), sigma, r, T, q)))
                gapd = S1 * (1 - dx) - K - cd - (S1 * (1 - dx) / q2) * (
                        1 - np.exp(-q * T) * norm.cdf(bsm_d1(K, S1 * (1 - dx), sigma, r, T, q)))
                dfx = (gapu - gapd)/(2 * dx * S1)

                if np.abs(gap0 / K) <= epsilon:
                    break
                else:
                    S1 -= gap0/dfx
                    i += 1
            if F >= S1:
                Call = F - K
            else:
                bsmc = bsm(K, F, sigma, r, T, q, 'call')
                Call = bsmc + (S1 / q2) * (
                        1 - np.exp(-q * T) * norm.cdf(bsm_d1(K, S1, sigma, r, T, q))
                ) * (F / S1)**q2
        return Call

    elif direction.lower() == 'put':
        S2 = K
        while i <= imax:
            p0 = bsm(K, S2, sigma, r, T, q, 'put')
            pu = bsm(K, S2 * (1 + dx), sigma, r, T, q, 'put')
            pd = bsm(K, S2 * (1 - dx), sigma, r, T, q, 'put')
            gap0 = K - S2 - p0 + (S2 / q1) * (1 - np.exp(-q * T) * norm.cdf(-bsm_d1(K, S2, sigma, r, T, q)))
            gapu = K - S2 * (1 + dx) - pu + (S2 * (1 + dx) / q1) * (
                    1 - np.exp(-q * T) * norm.cdf(-bsm_d1(K, S2 * (1 + dx), sigma, r, T, q)))
            gapd = K - S2 * (1 - dx) - pd + (S2 * (1 - dx) / q1) * (
                    1 - np.exp(-q * T) * norm.cdf(-bsm_d1(K, S2 * (1 - dx), sigma, r, T, q)))
            dfx = (gapu - gapd)/(2 * dx * S2)
            if np.abs(gap0) <= epsilon:
                break
            else:
                S2 -= gap0/dfx
                i += 1
        if F <= S2:
            Put = K - F
        else:
            bsmp = bsm(K, F, sigma, r, T, q, 'put')
            Put = bsmp - (S2 / q1) * (
                    1 - np.exp(-q * T) * norm.cdf(-bsm_d1(K, S2, sigma, r, T, q))
            ) * (F / S2)**q1
        return Put
    else:
        raise ValueError("parameter direction should either be 'call' or 'put'")


def bsm_imp_vol(K, F, sigma_guess, r, T, q, direction, opt):
    imax = 50
    i = 1
    epsilon = 0.00001
    while i <= imax:
        if np.abs(bsm(K, F, sigma_guess, r, T, q, direction) - opt) <= epsilon:
            break
        else:
            sigma_guess -= (
                                   bsm(K, F, sigma_guess, r, T, q, direction) - opt
                           ) / bsm_vega(K, F, sigma_guess, r, T, q)
            i += 1
    return sigma_guess


def baw_imp_vol(K, F, sigma_guess, r, T, q, direction, epsilon, opt):
    imax = 200
    i = 0
    # epsilon = 0.00001
    dx = 0.0001
    while i <= imax:
        if np.abs(baw(K, F, sigma_guess, r, T, q, direction, epsilon) - opt) <= epsilon:
            break
        else:
            dfx = (
                          baw(
                              K, F, sigma_guess * (1 + dx), r, T, q, direction, epsilon
                          ) - baw(
                      K, F, sigma_guess * (1 - dx), r, T, q, direction, epsilon
                  )
                  ) / (2 * sigma_guess * dx)
            sigma_guess -= (
                                   baw(K, F, sigma_guess, r, T, q, direction, epsilon) - opt
                           ) / dfx
            i += 1
    return sigma_guess


def crr_convertible_bond(K, F, sigma, T, steps, r=0.015, bond=100):
    delta_t = T / steps

    u = np.exp(sigma * np.sqrt(delta_t))
    d = 1 / u
    p = (np.exp(r * delta_t) - d) / (u - d)

    lattice = np.zeros((steps + 1, steps + 1))

    for i in range(steps + 1):
        spot_value = F * u ** (steps - i) * d ** i
        lattice[i, steps] = max(bond / K * spot_value, bond)

    discount_factor = np.exp(-r * delta_t)
    for j in range(steps - 1, -1, -1):
        for i in range(j + 1):
            fu = lattice[i, j + 1]
            fd = lattice[i, j + 1]
            discount_value = discount_factor * ((p * fu) + (1 - p) * fd)

            spot_value = F * u ** (j - i) * d ** i
            convert_value = bond / K * spot_value
            if convert_value > discount_value:
                a = 1
            lattice[i, j] = max(discount_value, convert_value)

    return lattice[0, 0]


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    ac = []
    ad = []
    ag = []
    x_axis = np.array(range(0, 100))
    for i in x_axis:
        print(i)
        # i = 80
        ac.append(
            bsm(
                K=50,
                F=i,
                sigma=0.5,
                T=1,
                q=0,
                r=0.015,
                direction='call'
            )
        )
        ad.append(
            bsm_delta(
                K=50,
                F=i,
                sigma=0.3,
                T=1,
                q=0,
                r=0.015,
                direction='call'
            )
        )
        ag.append(
            bsm_gamma(
                K=50,
                F=i,
                sigma=0.3,
                T=1,
                q=0,
                r=0.015,
            )
        )

    df = pd.DataFrame(
        {
            'spot': x_axis,
            'price': ac,
            'delta': ad,
            'gamma': ag
        }
    )
    df.to_excel('option_price.xlsx')

    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)
    ax1.plot(x_axis, ac)

    plt.show()









#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @FileName  :QueryException.py
# @Time      :2021/4/25 4:59 下午
# @Author    :meng.lv


class QueryException(BaseException):
    def __init__(self, data):
        self.data = data

    def __str__(self):
        return "%s : %s" % (self.data['returnCode'], self.data['returnMsg'])
